import json
import os
import platform

import pytest

from conan.api.model import RecipeReference, PkgReference
from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient


@pytest.fixture(scope="module")
def created_package():
    t = TestClient(light=True)
    t.save({"conanfile.py": GenConanfile()})
    t.run("create . --name foo --version 1.0")
    recipe_layout = t.exported_layout()
    pkg_layout = t.created_layout()
    return t, recipe_layout, pkg_layout


def test_cache_path(created_package):
    t, recipe_layout, pkg_layout = created_package

    # Basic recipe paths, without specifying revision

    recipe_revision = recipe_layout.reference.revision
    pref = pkg_layout.reference
    t.run("cache path foo/1.0")
    assert recipe_layout.export() == str(t.out).rstrip()
    t.run("cache path foo/1.0#latest")
    assert recipe_layout.export() == str(t.out).rstrip()
    t.run("cache path foo/1.0 --folder=export_source")
    assert recipe_layout.export_sources() == str(t.out).rstrip()
    t.run("cache path foo/1.0 --folder=source")
    assert recipe_layout.source() == str(t.out).rstrip()

    # Basic recipe paths, with revision
    t.run(f"cache path foo/1.0#{recipe_revision}")
    assert recipe_layout.export() == str(t.out).rstrip()
    t.run(f"cache path foo/1.0#{recipe_revision} --folder=export_source")
    assert recipe_layout.export_sources() == str(t.out).rstrip()
    t.run(f"cache path foo/1.0#{recipe_revision} --folder=source")
    assert recipe_layout.source() == str(t.out).rstrip()

    # Basic package paths, without specifying revision
    t.run(f"cache path foo/1.0:{pref.package_id}")
    assert pkg_layout.package() == str(t.out).rstrip()
    t.run(f"cache path foo/1.0:{pref.package_id} --folder=build")
    assert pkg_layout.build() == str(t.out).rstrip()

    # Basic package paths, with recipe-revision
    t.run(f"cache path foo/1.0#{recipe_revision}:{pref.package_id}")
    assert pkg_layout.package() == str(t.out).rstrip()
    t.run(f"cache path foo/1.0#{recipe_revision}:{pref.package_id}#latest")
    assert pkg_layout.package() == str(t.out).rstrip()
    t.run(f"cache path foo/1.0#{recipe_revision}:{pref.package_id} --folder=build")
    assert pkg_layout.build() == str(t.out).rstrip()

    # Basic package paths, with both revisions
    t.run(f"cache path foo/1.0#{recipe_revision}:{pref.package_id}#{pref.revision}")
    assert pkg_layout.package() == str(t.out).rstrip()
    t.run(f"cache path foo/1.0#{recipe_revision}:{pref.package_id}#{pref.revision} --folder=build")
    assert pkg_layout.build() == str(t.out).rstrip()


def test_cache_path_exist_errors(created_package):
    t, recipe_layout, pkg_layout = created_package
    recipe_revision = recipe_layout.reference.revision
    pref = pkg_layout.reference

    t.run("cache path nonexist/1.0", assert_error=True)
    assert "ERROR: Recipe 'nonexist/1.0' not found" in t.out

    t.run("cache path nonexist/1.0#rev", assert_error=True)
    assert "ERROR: Recipe 'nonexist/1.0#rev' not found" in t.out

    t.run("cache path foo/1.0#rev", assert_error=True)
    # TODO: Improve this error message
    assert "ERROR: Recipe 'foo/1.0#rev' not found" in t.out

    t.run(f"cache path foo/1.0:pid1", assert_error=True)
    assert f"ERROR: 'foo/1.0#{recipe_revision}:pid1' not found in cache" in t.out

    t.run(f"cache path foo/1.0#{recipe_revision}:pid1", assert_error=True)
    assert f"ERROR: 'foo/1.0#{recipe_revision}:pid1' not found in cache" in t.out

    t.run(f"cache path foo/1.0#{recipe_revision}:{pref.package_id}#rev2", assert_error=True)
    assert f"ERROR: No entry for package 'foo/1.0#{recipe_revision}:{pref.package_id}#rev2" in t.out


def test_cache_path_arg_errors():
    t = TestClient()
    # build, cannot obtain build without pref
    t.run("cache path foo/1.0 --folder build", assert_error=True)
    assert "ERROR: '--folder build' requires a valid package reference" in t.out

    # Invalid reference
    t.run("cache path patata", assert_error=True)
    assert "ERROR: patata is not a valid recipe reference" in t.out

    # source, cannot obtain build without pref
    t.run("cache path foo/1.0:pid --folder source", assert_error=True)
    assert "ERROR: '--folder source' requires a recipe reference" in t.out


def test_cache_path_does_not_exist_folder():
    client = TestClient(default_server_user=True)
    conanfile = GenConanfile()
    client.save({"conanfile.py": conanfile})
    client.run("create . --name=mypkg --version=0.1")
    pref = client.created_package_reference("mypkg/0.1")
    client.run("upload * --confirm -r default")
    client.run("remove * -c")

    client.run(f"install --requires mypkg/0.1")
    client.run(f"cache path {pref} --folder build", assert_error=True)
    assert f"ERROR: 'build' folder does not exist for the reference {pref}" in client.out


def test_cache_path_output_json():
    client = TestClient()
    conanfile = GenConanfile("mypkg", "0.1")
    client.save({"conanfile.py": conanfile})
    client.run("export .")
    layout = client.exported_layout()
    client.run("cache path mypkg/0.1 --format=json")
    output = json.loads(client.stdout)
    assert output == {"cache_path": os.path.join(layout.base_folder, "e")}


class TestCacheRef:

    def test_cache_path(self, created_package):
        t, recipe_layout, pkg_layout = created_package

        t.run(f'cache ref "{os.path.dirname(recipe_layout.export())}"')
        assert f"foo/1.0#{recipe_layout.reference.revision}" in t.out

        t.run(f'cache ref "{recipe_layout.export()}"')
        assert f"foo/1.0#{recipe_layout.reference.revision}" in t.out

        t.run(f'cache ref "{recipe_layout.source()}"')
        assert f"foo/1.0#{recipe_layout.reference.revision}" in t.out

        t.run(f'cache ref "{os.path.dirname(pkg_layout.package())}"')
        assert f"foo/1.0#{recipe_layout.reference.revision}" in t.out

        t.run(f'cache ref "{pkg_layout.build()}"')
        assert f"foo/1.0#{recipe_layout.reference.revision}" in t.out

        t.run(f'cache ref "{pkg_layout.package()}"')
        assert f"foo/1.0#{recipe_layout.reference.revision}" in t.out

    def test_downloaded_paths(self):
        t = TestClient(light=True, default_server_user=True)
        t.save({"conanfile.py": GenConanfile()})
        t.run("create . --name foo --version 1.0")
        t.run("upload * -c -r=default")
        t.run("remove * -c")
        t.run("install --requires=foo/1.0")
        ref = RecipeReference.loads("foo/1.0#4d670581ccb765839f2239cc8dff8fbd")
        pref = PkgReference(ref, "da39a3ee5e6b4b0d3255bfef95601890afd80709")
        recipe_layout = t.get_latest_ref_layout(ref)
        pkg_layout = t.get_latest_pkg_layout(pref)

        t.run(f'cache ref "{os.path.dirname(recipe_layout.export())}"')
        assert f"foo/1.0#{recipe_layout.reference.revision}" in t.out

        t.run(f'cache ref "{recipe_layout.export()}"')
        assert f"foo/1.0#{recipe_layout.reference.revision}" in t.out

        t.run(f'cache ref "{os.path.dirname(pkg_layout.package())}"')
        assert f"foo/1.0#{recipe_layout.reference.revision}" in t.out

        t.run(f'cache ref "{pkg_layout.package()}"')
        assert f"foo/1.0#{recipe_layout.reference.revision}" in t.out

    def test_errors(self):
        t = TestClient(light=True)
        t.run('cache ref "/some/non/existing/path"', assert_error=True)
        assert "ERROR: Reference for this path not found in cache" in t.out

        if platform.system() == "Windows":
            t.run("cache ref J:/not/path", assert_error=True)
            assert "ERROR: Invalid path: J:/not/path" in t.out



import os
import platform
from unittest.mock import MagicMock

import pytest

from conan.test.utils.mocks import ConanFileMock, RedirectedTestOutput
from conan.test.utils.test_files import temp_folder
from conan.test.utils.tools import redirect_output
from conan.tools.files import save
from conan.internal.model.cpp_info import CppInfo
from conan.internal.model.pkg_type import PackageType
from conan.api.model import RecipeReference
from conan.internal.util.files import save


@pytest.fixture
def conanfile():
    c = ConanFileMock()
    c._conan_node = MagicMock(ref=RecipeReference(""))
    return c


@pytest.mark.parametrize("lib_name, libs", [
    ("myliblin.a", ["myliblin"]),
    ("libmyliblin.a", ["myliblin"]),
    ("mylibmac.a", ["mylibmac"]),
    ("mylibwin.lib", ["mylibwin"]),
    ("libmylibwin.lib", ["libmylibwin"]),
    ("mylibwin2.if.lib", ["mylibwin2.if.lib"]),
    ("mylibwin2.if.lib", ["mylibwin2"])
])
def test_simple_deduce_locations_static(lib_name, libs, conanfile):
    folder = temp_folder()
    location = os.path.join(folder, "libdir", lib_name)
    save(location, "")

    cppinfo = CppInfo()
    cppinfo.libdirs = ["libdir"]
    cppinfo.libs = libs
    cppinfo.set_relative_base_folder(folder)

    result = cppinfo.deduce_full_cpp_info(conanfile)
    assert result.location == location.replace("\\", "/")
    assert result.link_location is None
    assert result.type == "static-library"


def test_deduce_shared_link_locations(conanfile):
    folder = temp_folder()
    imp_location = os.path.join(folder, "libdir", "mylib.lib")
    save(imp_location, "")
    location = os.path.join(folder, "bindir", "mylib.dll")
    save(location, "")

    cppinfo = CppInfo()
    cppinfo.libdirs = ["libdir"]
    cppinfo.bindirs = ["bindir"]
    cppinfo.libs = ["mylib"]
    cppinfo.set_relative_base_folder(folder)

    result = cppinfo.deduce_full_cpp_info(conanfile)
    assert result.location == location.replace("\\", "/")
    assert result.link_location == imp_location.replace("\\", "/")
    assert result.type == "shared-library"


@pytest.mark.parametrize("lib_name, libs", [
    ("liblog4cxx.so.15.2.0", ["log4cxx"]),
    ("libapr-1.0.dylib", ["apr-1"]),
    ("libapr-1.so.0.7.4", ["apr-1"]),
    ("libgrpc++_alts.so.1.67.1", ["grpc++_alts"])
])
def test_complex_deduce_locations_shared(lib_name, libs, conanfile):
    """
    Tests real examples of shared library names in Linux/MacOS,
    e.g., log4cxx, apr-1, etc.

    Related issue: https://github.com/conan-io/conan/issues/16990
    """
    folder = temp_folder()
    location = os.path.join(folder, "libdir", lib_name)
    save(location, "")

    cppinfo = CppInfo()
    cppinfo.libdirs = ["libdir"]
    cppinfo.libs = libs
    cppinfo.set_relative_base_folder(folder)

    result = cppinfo.deduce_full_cpp_info(conanfile)
    assert result.location == location.replace("\\", "/")
    assert result.link_location is None
    assert result.type == "shared-library"


@pytest.mark.parametrize("lib_name, dll_name, libs, pkg_name", [
    ("libcurl_imp.lib", "libcurl.dll", ["libcurl_imp"], "libcurl"),
    ("libcrypto.lib", "libcrypto-3-x64.dll", ["libcrypto"], "crypto"),
    ("libssl.lib", "libssl-3-x64.dll", ["libssl"], "ssl"),
    ("zdll.lib", "zlib1.dll", ["zdll"], "zlib")
])
def test_windows_shared_link_locations(lib_name, dll_name, libs, pkg_name, conanfile):
    """
    Tests real examples of shared library names in Windows,
    e.g., openssl, zlib, libcurlb, etc.
    """
    folder = temp_folder()
    imp_location = os.path.join(folder, "libdir", lib_name)
    save(imp_location, "")
    location = os.path.join(folder, "bindir", dll_name)
    save(location, "")

    cppinfo = CppInfo()
    cppinfo.libdirs = ["libdir"]
    cppinfo.bindirs = ["bindir"]
    cppinfo.libs = libs
    cppinfo.set_relative_base_folder(folder)

    conanfile._conan_node.ref = RecipeReference(name=pkg_name)
    result = cppinfo.deduce_full_cpp_info(conanfile)
    assert result.location == location.replace("\\", "/")
    assert result.link_location == imp_location.replace("\\", "/")
    assert result.type == "shared-library"


@pytest.mark.parametrize("lib_info", [
    {"charset": ["charset.lib", "charset-1.dll"],
     "iconv": ["iconv.lib", "iconv-2.dll"]},
    {"charset": ["libcharset.so.1.0.0"],
     "iconv": ["libiconv.so.2.6.1"]},
])
def test_windows_several_shared_link_locations(lib_info, conanfile):
    """
    Tests a real model as LIBICONV with several libs defined in the root component
    """
    folder = temp_folder()
    locations = {}
    is_windows = False
    for lib_name, lib_files in lib_info.items():
        imp_location = os.path.join(folder, "libdir", lib_files[0])
        save(imp_location, "")
        if len(lib_files) > 1:
            is_windows = True
            location = os.path.join(folder, "bindir", lib_files[1])
            save(location, "")
        else:
            location = imp_location  # Linux
        locations[lib_name] = (location, imp_location)

    cppinfo = CppInfo()
    cppinfo.libdirs = ["libdir"]
    cppinfo.bindirs = ["bindir"]
    cppinfo.libs = list(lib_info.keys())
    cppinfo.set_relative_base_folder(folder)

    result = cppinfo.deduce_full_cpp_info(conanfile)
    for lib_name in lib_info:
        assert result.components[f"_{lib_name}"].location == locations[lib_name][0].replace("\\", "/")
        if is_windows:
            assert result.components[f"_{lib_name}"].link_location == locations[lib_name][1].replace("\\", "/")
        assert result.components[f"_{lib_name}"].type == "shared-library"


@pytest.mark.parametrize("lib, symlinks", [
    # symlinks == "real_file <- symlink1 <- symlink2 <- ... <- symlinkN"
    # Issue related: https://github.com/conan-io/conan/issues/17417
    ("png", "libpng16.so.16.44.0 <- libpng16.so.16 <- libpng16.so <- libpng.so")
])
@pytest.mark.skipif(platform.system() == "Windows", reason="Can't apply symlink on Windows")
def test_shared_link_locations_symlinks(lib, symlinks, conanfile):
    """
    Tests auto deduce location is not resolving symlinks by default. For instance:
        .
        ├── libpng.so -> libpng16.so  (exact match)
        ├── libpng16.so -> libpng16.so.16
        ├── libpng16.so.16 -> libpng16.so.16.44.0
        └── libpng16.so.16.44.0  (real one)


    Issues related:
        - https://github.com/conan-io/conan/issues/17417
        - https://github.com/conan-io/conan/issues/17721
    """
    folder = temp_folder()
    all_files = symlinks.split(" <- ")  # [real_one, sym1, sym2, ...]
    # Real one (first item from list)
    real_location = os.path.join(folder, "libdir", all_files.pop(0))
    save(real_location, "")
    # Symlinks
    prev_path = real_location
    for file in all_files:
        sym = os.path.join(folder, "libdir", file)
        os.symlink(prev_path, sym)
        prev_path = sym
    # Exact match and symlink (latest item from list)
    exact_match = os.path.join(folder, "libdir", all_files[-1])

    cppinfo = CppInfo()
    cppinfo.libdirs = ["libdir"]
    cppinfo.libs = [lib]
    cppinfo.set_relative_base_folder(folder)

    result = cppinfo.deduce_full_cpp_info(conanfile)
    assert result.location == exact_match
    assert result.type == "shared-library"


@pytest.mark.parametrize("static", [True, False])
def test_error_if_shared_and_static_found(static, conanfile):
    folder = temp_folder()
    save(os.path.join(folder, "libdir", "libmylib.a"), "")
    save(os.path.join(folder, "libdir", "libmylib.so"), "")

    cppinfo = CppInfo()
    cppinfo.libdirs = ["libdir"]
    cppinfo.libs = ["mylib"]
    cppinfo.set_relative_base_folder(folder)
    folder = folder.replace("\\", "/")
    if static:
        conanfile.package_type = PackageType.STATIC
    result = cppinfo.deduce_full_cpp_info(conanfile)
    ext = "a" if static else "so"
    assert result.location == f"{folder}/libdir/libmylib.{ext}"
    assert result.type == (PackageType.STATIC if static else PackageType.SHARED)


def test_warning_windows_if_more_than_one_dll(conanfile):
    folder = temp_folder()
    save(os.path.join(folder, "libdir", "mylib.a"), "")
    save(os.path.join(folder, "bindir", "libx.dll"), "")
    save(os.path.join(folder, "bindir", "liby.dll"), "")

    cppinfo = CppInfo()
    cppinfo.libdirs = ["libdir"]
    cppinfo.bindirs = ["bindir"]
    cppinfo.libs = ["mylib"]
    cppinfo.set_relative_base_folder(folder)
    folder = folder.replace("\\", "/")
    output = RedirectedTestOutput()  # Initialize each command
    with redirect_output(output):
        result = cppinfo.deduce_full_cpp_info(conanfile)
    assert "WARN: There were several matches for Lib mylib" in output
    assert result.location == f"{folder}/bindir/libx.dll"
    assert result.type == "shared-library"


@pytest.mark.parametrize("prefix", [True, False])
def test_multiple_matches_exact_match(prefix, conanfile):
    # If the match is perfect, do not warn
    folder = temp_folder()
    prefix = "lib" if prefix else ""
    save(os.path.join(folder, "libdir", f"{prefix}mylib.a"), "")
    save(os.path.join(folder, "libdir", f"{prefix}mylib_imp.a"), "")
    save(os.path.join(folder, "libdir", f"{prefix}mylib_other.a"), "")

    cppinfo = CppInfo()
    cppinfo.libdirs = ["libdir"]
    cppinfo.libs = ["mylib"]
    cppinfo.set_relative_base_folder(folder)
    folder = folder.replace("\\", "/")
    output = RedirectedTestOutput()  # Initialize each command
    with redirect_output(output):
        result = cppinfo.deduce_full_cpp_info(conanfile)
    assert "WARN: There were several matches for Lib mylib" not in output
    assert result.location == f"{folder}/libdir/{prefix}mylib.a"
    assert result.type == "static-library"



@pytest.mark.parametrize("lib_name, libs", [
    ("harfbuzz", ["harfbuzz-icu.lib", "harfbuzz.lib"]),
])
def test_several_libs_and_exact_match(lib_name, libs, conanfile):
    """
    Testing that we're keeping the exact match at first instead the similar one

    Issue related: https://github.com/conan-io/conan/issues/17974
    """
    folder = temp_folder()
    for lib in libs:
        save(os.path.join(folder, "libdir", lib), "")

    cppinfo = CppInfo()
    cppinfo.libdirs = ["libdir"]
    cppinfo.libs = [lib_name]
    cppinfo.set_relative_base_folder(folder)
    folder = folder.replace("\\", "/")
    result = cppinfo.deduce_full_cpp_info(conanfile)
    assert result.location == f"{folder}/libdir/{lib_name}.lib"
    assert result.type == "static-library"


def test_sources(conanfile):
    folder = temp_folder()
    save(os.path.join(folder, "src", "mylib.cpp"), "")
    cppinfo = CppInfo()
    cppinfo.sources = ["src/mylib.cpp"]
    cppinfo.set_relative_base_folder(folder)
    result = cppinfo.deduce_full_cpp_info(conanfile)
    assert result.location is None
    assert result.type == "header-library"



from collections import OrderedDict

from conan.test.utils.tools import TestServer, GenConanfile, TestClient


def test_cascade():
    """
    app -> E -> D -> B -> A
      \\-> F -> C -------/
    """
    server = TestServer()
    servers = OrderedDict([("default", server)])
    c = TestClient(servers=servers)
    c.save({"a/conanfile.py": GenConanfile("liba", "1.0"),
            "b/conanfile.py": GenConanfile("libb", "1.0").with_requires("liba/1.0"),
            "c/conanfile.py": GenConanfile("libc", "1.0").with_requires("liba/1.0"),
            "d/conanfile.py": GenConanfile("libd", "1.0").with_requires("libb/1.0"),
            "e/conanfile.py": GenConanfile("libe", "1.0").with_requires("libd/1.0"),
            "f/conanfile.py": GenConanfile("libf", "1.0").with_requires("libc/1.0", "libd/1.0"),
            "app/conanfile.py": GenConanfile().with_requires("libe/1.0", "libf/1.0")})

    for pkg in ("a", "b", "c", "d", "e", "f"):
        c.run(f"create {pkg}")

    def _assert_built(refs):
        for ref in refs:
            assert "{}: Copying sources to build folder".format(ref) in c.out
        for ref in ["liba/1.0", "libb/1.0", "libc/1.0", "libd/1.0", "libe/1.0", "libf/1.0"]:
            if ref not in refs:
                assert "{}: Copying sources to build folder".format(ref) not in c.out

    # Building A everything is built
    c.run("install app --build=liba* --build cascade")
    _assert_built(["liba/1.0", "libb/1.0", "libc/1.0", "libd/1.0", "libe/1.0", "libf/1.0"])

    c.run("install app --build=libd* --build cascade")
    _assert_built(["libd/1.0", "libe/1.0", "libf/1.0"])

    c.run("install app --build=libe* --build cascade")
    _assert_built(["libe/1.0"])

    c.run("install app  --build cascade")
    _assert_built([])

    c.run("install app --build=libc* --build cascade")
    _assert_built(["libc/1.0", "libf/1.0"])



import textwrap

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient


def test_relocatable_toolchain():
    """ Implements the following use case:
    - base/1.0 implements an SDK that needs to be relocated in every maching, but this package
      contains the non relocatable part and the relocation scripts
    - sdk/1.0 --tool_requires-> base/1.0 and implements just the relocation execution, copying
      whatever is necessary from base, and relocating it.
      It defines build_policy = "missing" and upload policy = "skip"
    https://github.com/conan-io/conan/issues/5059
    """
    c = TestClient(default_server_user=True)
    base = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.files import save, copy
        class Pkg(ConanFile):
            name = "base"
            version = "1.0"
            settings = "arch"

            def build(self):
                save(self, "sdk.txt", f"arch:{self.settings.arch}=>{self.settings_target.arch}")

            def package(self):
                copy(self, "*", self.build_folder, self.package_folder)
        """)
    sdk = textwrap.dedent("""
        import os
        from conan import ConanFile
        from conan.tools.files import load, copy, save
        class Pkg(ConanFile):
            name = "sdk"
            version = "1.0"
            settings = "arch"
            build_policy = "missing"
            upload_policy = "skip"

            def build_requirements(self):
                self.tool_requires("base/1.0")

            def package(self):
                # Whatever modification, customization, RPATHs, symlinks, etc
                pkg_folder = self.dependencies.build["base"].package_folder
                sdk = load(self, os.path.join(pkg_folder, "sdk.txt"))
                save(self, os.path.join(self.package_folder, "sdk.txt"), "CUSTOM PATH: " + sdk)

            def package_info(self):
                self.output.info(f"SDK INFO: {load(self, 'sdk.txt')}!!!")
        """)
    c.save({"linux": "[settings]\nos=Linux\narch=x86_64",
            "embedded": "[settings]\narch=armv8",
            "base/conanfile.py": base,
            "sdk/conanfile.py": sdk,
            "consumer/conanfile.py": GenConanfile("app", "1.0").with_tool_requires("sdk/1.0")})
    c.run("create base -pr:h=embedded -pr:b=linux --build-require")
    c.run("export sdk")
    c.run("install consumer -pr:h=embedded -pr:b=linux")
    c.assert_listed_binary({"base/1.0": ("62e589af96a19807968167026d906e63ed4de1f5", "Cache"),
                            "sdk/1.0": ("62e589af96a19807968167026d906e63ed4de1f5", "Build")},
                           build=True)
    assert "sdk/1.0: Calling package()" in c.out
    assert "sdk/1.0: SDK INFO: CUSTOM PATH: arch:x86_64=>armv8!!!" in c.out

    c.run("install consumer -pr:h=embedded -pr:b=linux -v")
    c.assert_listed_binary({"base/1.0": ("62e589af96a19807968167026d906e63ed4de1f5", "Skip"),
                            "sdk/1.0": ("62e589af96a19807968167026d906e63ed4de1f5", "Cache")},
                           build=True)
    assert "sdk/1.0: Calling package()" not in c.out
    assert "sdk/1.0: SDK INFO: CUSTOM PATH: arch:x86_64=>armv8!!!" in c.out

    # If I upload everything and remove:
    c.run("upload * -r=default -c")
    c.run("remove * -c")
    c.run("install consumer -pr:h=embedded -pr:b=linux")
    c.assert_listed_binary(
        {"base/1.0": ("62e589af96a19807968167026d906e63ed4de1f5", "Download (default)"),
         "sdk/1.0": ("62e589af96a19807968167026d906e63ed4de1f5", "Build")},
        build=True)
    assert "sdk/1.0: Calling package()" in c.out
    assert "sdk/1.0: SDK INFO: CUSTOM PATH: arch:x86_64=>armv8!!!" in c.out

    # we can even remove the binary!
    c.run("remove base/1.0:* -c")
    c.run("install consumer -pr:h=embedded -pr:b=linux -v")
    c.assert_listed_binary({"base/1.0": ("62e589af96a19807968167026d906e63ed4de1f5", "Skip"),
                            "sdk/1.0": ("62e589af96a19807968167026d906e63ed4de1f5", "Cache")},
                           build=True)
    assert "sdk/1.0: SDK INFO: CUSTOM PATH: arch:x86_64=>armv8!!!" in c.out



import os
import shutil
import textwrap

import pytest

from conan.test.assets.sources import gen_function_h, gen_function_cpp
from conan.test.utils.test_files import temp_folder
from conan.test.utils.tools import TestClient


@pytest.fixture(scope="session")
def _matrix_client():
    """
    matrix/1.0, just static, no test_package
    """
    c = TestClient()
    c.run("new cmake_lib -d name=matrix -d version=1.0")
    c.run("create . -tf=")
    return c


@pytest.fixture(scope="session")
def _matrix_client_shared(_matrix_client):
    _matrix_client.run("create . -o *:shared=True -tf=")
    return _matrix_client


@pytest.fixture(scope="session")
def _matrix_client_debug(_matrix_client):
    _matrix_client.run("create . -s build_type=Debug -tf=")
    return _matrix_client


@pytest.fixture()
def matrix_client(_matrix_client):
    c = TestClient()
    c.cache_folder = os.path.join(temp_folder(), ".conan2")
    shutil.copytree(_matrix_client.cache_folder, c.cache_folder)
    return c


@pytest.fixture()
def matrix_client_nospace(_matrix_client):
    c = TestClient(path_with_spaces=False)
    c.cache_folder = os.path.join(temp_folder(path_with_spaces=False), ".conan2")
    shutil.copytree(_matrix_client.cache_folder, c.cache_folder)
    return c


@pytest.fixture()
def matrix_client_shared(_matrix_client_shared):
    c = TestClient()
    c.cache_folder = os.path.join(temp_folder(), ".conan2")
    shutil.copytree(_matrix_client_shared.cache_folder, c.cache_folder)
    return c


@pytest.fixture()
def matrix_client_shared_debug(_matrix_client_shared, _matrix_client_debug):
    c = TestClient()
    c.cache_folder = os.path.join(temp_folder(), ".conan2")
    shutil.copytree(_matrix_client_shared.cache_folder, c.cache_folder)
    return c


@pytest.fixture()
def matrix_client_debug(_matrix_client_debug):
    c = TestClient()
    c.cache_folder = os.path.join(temp_folder(), ".conan2")
    shutil.copytree(_matrix_client_debug.cache_folder, c.cache_folder)
    return c


@pytest.fixture(scope="session")
def _transitive_libraries(_matrix_client):
    """
    engine/1.0->matrix/1.0
    """
    c = TestClient()
    c.cache_folder = os.path.join(temp_folder(), ".conan2")
    shutil.copytree(_matrix_client.cache_folder, c.cache_folder)
    c.save({}, clean_first=True)
    c.run("new cmake_lib -d name=engine -d version=1.0 -d requires=matrix/1.0")
    # create both static and shared
    c.run("create . -tf=")
    c.run("create . -o engine/*:shared=True -tf=")
    return c


@pytest.fixture()
def transitive_libraries(_transitive_libraries):
    c = TestClient()
    c.cache_folder = os.path.join(temp_folder(), ".conan2")
    shutil.copytree(_transitive_libraries.cache_folder, c.cache_folder)
    return c


@pytest.fixture(scope="session")
def _matrix_client_components():
    """
    2 components, different than the package name
    """
    c = TestClient()
    headers_h = textwrap.dedent("""
        #include <iostream>
        #ifndef MY_MATRIX_HEADERS_DEFINE
        #error "Fatal error MY_MATRIX_HEADERS_DEFINE not defined"
        #endif
        void headers(){ std::cout << "Matrix headers: Release!" << std::endl;
            #if __cplusplus
            std::cout << "  Matrix headers __cplusplus: __cplusplus" << __cplusplus << std::endl;
            #endif
        }
        """)
    vector_h = gen_function_h(name="vector")
    vector_cpp = gen_function_cpp(name="vector", includes=["vector"])
    module_h = gen_function_h(name="module")
    module_cpp = gen_function_cpp(name="module", includes=["module", "vector"], calls=["vector"])

    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.cmake import CMake

        class Matrix(ConanFile):
            name = "matrix"
            version = "1.0"
            settings = "os", "compiler", "build_type", "arch"
            generators = "CMakeToolchain"
            exports_sources = "src/*", "CMakeLists.txt"

            def build(self):
                cmake = CMake(self)
                cmake.configure()
                cmake.build()

            def package(self):
                cmake = CMake(self)
                cmake.install()

            def package_info(self):
                self.cpp_info.default_components = ["vector", "module"]

                self.cpp_info.components["headers"].includedirs = ["include/headers"]
                self.cpp_info.components["headers"].set_property("cmake_target_name", "MatrixHeaders")
                self.cpp_info.components["headers"].defines = ["MY_MATRIX_HEADERS_DEFINE=1"]
                # Few flags to cover that CMakeDeps doesn't crash with them
                if self.settings.compiler == "msvc":
                    self.cpp_info.components["headers"].cxxflags = ["/Zc:__cplusplus"]
                    self.cpp_info.components["headers"].cflags = ["/Zc:__cplusplus"]
                    self.cpp_info.components["headers"].system_libs = ["ws2_32"]
                else:
                    self.cpp_info.components["headers"].system_libs = ["m"]
                    # Just to verify CMake don't break
                    self.cpp_info.sharedlinkflags = ["-z now", "-z relro"]
                    self.cpp_info.exelinkflags = ["-z now", "-z relro"]

                self.cpp_info.components["vector"].libs = ["vector"]
                self.cpp_info.components["vector"].includedirs = ["include"]
                self.cpp_info.components["vector"].libdirs = ["lib"]

                self.cpp_info.components["module"].libs = ["module"]
                self.cpp_info.components["module"].includedirs = ["include"]
                self.cpp_info.components["module"].libdirs = ["lib"]
                self.cpp_info.components["module"].requires = ["vector"]
          """)

    cmakelists = textwrap.dedent("""
       set(CMAKE_CXX_COMPILER_WORKS 1)
       set(CMAKE_CXX_ABI_COMPILED 1)
       cmake_minimum_required(VERSION 3.15)
       project(matrix CXX)

       add_library(vector src/vector.cpp)
       add_library(module src/module.cpp)
       add_library(headers INTERFACE)
       target_link_libraries(module PRIVATE vector)

       set_target_properties(headers PROPERTIES PUBLIC_HEADER "src/headers.h")
       set_target_properties(module PROPERTIES PUBLIC_HEADER "src/module.h")
       set_target_properties(vector PROPERTIES PUBLIC_HEADER "src/vector.h")
       install(TARGETS vector module)
       install(TARGETS headers PUBLIC_HEADER DESTINATION include/headers)
       """)
    c.save({"src/headers.h": headers_h,
            "src/vector.h": vector_h,
            "src/vector.cpp": vector_cpp,
            "src/module.h": module_h,
            "src/module.cpp": module_cpp,
            "CMakeLists.txt": cmakelists,
            "conanfile.py": conanfile})
    c.run("create .")
    return c


@pytest.fixture()
def matrix_client_components(_matrix_client_components):
    c = TestClient()
    c.cache_folder = os.path.join(temp_folder(), ".conan2")
    shutil.copytree(_matrix_client_components.cache_folder, c.cache_folder)
    return c


@pytest.fixture(scope="session")
def _matrix_c_interface_client():
    c = TestClient()
    matrix_h = textwrap.dedent("""\
        #pragma once
        #ifdef __cplusplus
        extern "C" {
        #endif
            void matrix();
        #ifdef __cplusplus
        }
        #endif
        """)
    matrix_cpp = textwrap.dedent("""\
        #include "matrix.h"
        #include <iostream>
        #include <string>
        void matrix(){
            std::cout<< std::string("Hello Matrix!") <<std::endl;
        }
        """)
    # Having here the config.cmake code to be able to manually check what CMake generates
    cmake = textwrap.dedent("""\
        set(CMAKE_CXX_COMPILER_WORKS 1)
        set(CMAKE_CXX_ABI_COMPILED 1)
        cmake_minimum_required(VERSION 3.15)
        project(matrix C CXX)
        add_library(matrix STATIC src/matrix.cpp)
        target_include_directories(matrix PUBLIC
          $<BUILD_INTERFACE:${CMAKE_CURRENT_SOURCE_DIR}/include>
          $<INSTALL_INTERFACE:include>
        )
        set_target_properties(matrix PROPERTIES PUBLIC_HEADER "include/matrix.h")
        install(TARGETS matrix EXPORT matrixConfig)
        export(TARGETS matrix
            NAMESPACE matrix::
            FILE "${CMAKE_CURRENT_BINARY_DIR}/matrixConfig.cmake"
        )
        install(EXPORT matrixConfig
            DESTINATION "matrix/cmake"
            NAMESPACE matrix::
        )
        """)
    conanfile = textwrap.dedent("""\
        from conan import ConanFile
        from conan.tools.cmake import CMake, cmake_layout
        class Recipe(ConanFile):
            name = "matrix"
            version = "0.1"
            settings = "os", "compiler", "build_type", "arch"
            package_type = "static-library"
            generators = "CMakeToolchain"
            exports_sources = "CMakeLists.txt", "src/*", "include/*"
            languages = "C++"
            def build(self):
                cmake = CMake(self)
                cmake.configure()
                cmake.build()
            def layout(self):
                cmake_layout(self)
            def package(self):
                cmake = CMake(self)
                cmake.install()
            def package_info(self):
                self.cpp_info.libs = ["matrix"]
        """)
    c.save({"include/matrix.h": matrix_h,
            "src/matrix.cpp": matrix_cpp,
            "conanfile.py": conanfile,
            "CMakeLists.txt": cmake})
    c.run("create .")
    return c


@pytest.fixture()
def matrix_c_interface_client(_matrix_c_interface_client):
    c = TestClient()
    c.cache_folder = os.path.join(temp_folder(), ".conan2")
    shutil.copytree(_matrix_c_interface_client.cache_folder, c.cache_folder)
    return c



import json
import os
import textwrap

import pytest

from conan.api.conan_api import ConanAPI
from conan.test.assets.genconanfile import GenConanfile
from conan.internal.model.conf import BUILT_IN_CONFS
from conan.test.utils.test_files import temp_folder
from conan.test.utils.tools import TestClient
from conan.test.utils.env import environment_update


def test_missing_subarguments():
    """ config MUST run  with a subcommand. Otherwise, it MUST exits with error.
    """
    client = TestClient()
    client.run("config", assert_error=True)
    assert "ERROR: Exiting with code: 2" in client.out


class TestConfigHome:
    """ The test framework cannot test the CONAN_HOME env-var because it is not using it
    (it will break tests for maintainers that have the env-var defined)
    """
    def test_config_home_default(self):
        client = TestClient()
        client.run("config home")
        assert f"{client.cache_folder}\n" == client.stdout

        client.run("config home --format=text", assert_error=True)
        # It is not possible to use --format=text explicitly
        assert "--format=text" in client.out

    def test_api_uses_env_var_home(self):
        cache_folder = os.path.join(temp_folder(), "custom")
        with environment_update({"CONAN_HOME": cache_folder}):
            api = ConanAPI()
            assert api.cache_folder == cache_folder


def test_config_list():
    """
    'conan config list' shows all the built-in Conan configurations
    """
    client = TestClient()
    client.run("config list")
    for k, v in BUILT_IN_CONFS.items():
        assert f"{k}: {v}" in client.out
    client.run("config list --format=json")
    assert f"{json.dumps(BUILT_IN_CONFS, indent=4)}\n" == client.stdout

    client.run("config list cmake")
    assert "tools.cmake:cmake_program: Path to CMake executable" in client.out
    assert "core.download:parallel" not in client.out
    assert "tools.build:verbosity" not in client.out


def test_config_install():
    tc = TestClient()
    tc.save({'config/foo': ''})
    # This should not fail (insecure flag exists)
    tc.run("config install config --insecure")
    assert "foo" in os.listdir(tc.cache_folder)
    # Negative test, ensure we would be catching a missing arg if it did not exist
    tc.run("config install config --superinsecure", assert_error=True)


def test_config_install_conanignore():
    tc = TestClient()
    conanignore = textwrap.dedent("""
    a/*  # This is a tests
    b/c/*
    d/*
    tests/*
    # Next line is commented out, so it should be ignored
    # other_tests/*
    !b/c/important_file
    !b/c/important_folder/*
    """)
    tc.save({
        'config_folder/.conanignore': conanignore,
        'config_folder/a/test': '',
        'config_folder/abracadabra': '',
        'config_folder/b/bison': '',
        'config_folder/b/a/test2': '',
        'config_folder/b/c/helmet': '',
        'config_folder/b/c/important_file': '',
        'config_folder/b/c/important_folder/contents': '',
        'config_folder/d/prix': '',
        'config_folder/d/foo/bar': '',
        'config_folder/foo': '',
        'config_folder/tests/tester': '',
        'config_folder/other_tests/tester2': ''
    })

    def _assert_config_exists(path):
        assert os.path.exists(os.path.join(tc.cache_folder, path))

    def _assert_config_not_exists(path):
        assert not os.path.exists(os.path.join(tc.cache_folder, path))

    tc.run('config install config_folder')

    _assert_config_not_exists(".conanignore")

    _assert_config_not_exists("a")
    _assert_config_not_exists("a/test")

    _assert_config_exists("abracadabra")

    _assert_config_exists("b")
    _assert_config_exists("b/bison")
    _assert_config_exists("b/a/test2")
    _assert_config_not_exists("b/c/helmet")

    _assert_config_exists("b/c")
    _assert_config_exists("b/c/important_file")
    _assert_config_exists("b/c/important_folder/contents")

    _assert_config_not_exists("d/prix")
    _assert_config_not_exists("d/foo/bar")
    _assert_config_not_exists("d")

    _assert_config_exists("foo")

    _assert_config_not_exists("tests/tester")
    _assert_config_exists("other_tests/tester2")


def test_config_install_conanignore_ignore_all_allow_specific_workflow():
    tc = TestClient()
    conanignore = textwrap.dedent("""
    *
    !important_folder/*
    !important_file
    # We can even include the conanignore that we skip by default!
    !.conanignore
    """)
    tc.save({
        'config_folder/.conanignore': conanignore,
        'config_folder/a/test': '',
        'config_folder/abracadabra': '',
        'config_folder/important_folder/contents': '',
        'config_folder/important_file': '',
    })

    def _assert_config_exists(path):
        assert os.path.exists(os.path.join(tc.cache_folder, path))

    def _assert_config_not_exists(path):
        assert not os.path.exists(os.path.join(tc.cache_folder, path))

    tc.run('config install config_folder')

    _assert_config_exists(".conanignore")

    _assert_config_not_exists("a")
    _assert_config_not_exists("abracadabra")

    _assert_config_exists("important_folder/contents")
    _assert_config_exists("important_file")


@pytest.mark.parametrize("has_conanignore", [True, False])
@pytest.mark.parametrize("folder", [None, "myfolder", "myfolder/subfolder"])
def test_config_install_conanignore_walk_directories(has_conanignore, folder):
    tc = TestClient(light=True)
    if has_conanignore:
        conanignore = "*"
        tc.save({"config_folder/.conanignore": conanignore})
    tc.save({"config_folder/myfolder/subfolder/item.py": ""})

    folder_arg = f"-sf {folder} -tf {folder}" if folder else ""

    tc.run(f"config install config_folder {folder_arg}")
    if has_conanignore:
        assert not os.path.exists(os.path.join(tc.cache_folder, "myfolder", "subfolder", "item.py"))
    else:
        assert os.path.exists(os.path.join(tc.cache_folder, "myfolder", "subfolder", "item.py"))


def test_config_show():
    globalconf = textwrap.dedent("""
    tools.build:jobs=42
    tools.files.download:retry_wait=10
    tools.files.download:retry=7
    core.net.http:timeout=30
    core.net.http:max_retries=5
    zlib/*:user.mycategory:retry=True
    zlib/*:user.mycategory:foo=0
    zlib/*:user.myothercategory:foo=0
    """)
    tc = TestClient()
    tc.save_home({"global.conf": globalconf})
    tc.run("config show tools.build:jobs")
    assert "42" in tc.out

    tc.run("config show core*")
    assert "core.net.http:timeout" in tc.out
    assert "30" in tc.out
    assert "core.net.http:max_retries" in tc.out
    assert "5" in tc.out

    tc.run("config show *retr*")
    assert "tools.files.download:retry_wait" in tc.out
    assert "tools.files.download:retry" in tc.out
    assert "core.net.http:max_retries" in tc.out
    assert "zlib/*:user.mycategory:retry" in tc.out

    tc.run("config show zlib*")
    assert "zlib/*:user.mycategory:retry" in tc.out
    assert "zlib/*:user.mycategory:foo" in tc.out
    assert "zlib/*:user.myothercategory:foo" in tc.out

    tc.run("config show zlib/*")
    assert "zlib/*:user.mycategory:retry" in tc.out
    assert "zlib/*:user.mycategory:foo" in tc.out
    assert "zlib/*:user.myothercategory:foo" in tc.out

    tc.run("config show zlib/*:foo")
    assert "zlib/*:user.mycategory:foo" in tc.out
    assert "zlib/*:user.myothercategory:foo" in tc.out


@pytest.mark.parametrize("storage_path", [None, "p", "../foo"])
def test_config_clean(storage_path):
    tc = TestClient(light=True)
    absolut_storage_path = os.path.abspath(os.path.join(tc.current_folder, storage_path)) if storage_path else os.path.join(tc.cache_folder, "p")

    storage = f"core.cache:storage_path={storage_path}" if storage_path else ""
    tc.save_home({"global.conf": f"core.upload:retry=7\n{storage}",
                  "extensions/compatibility/mycomp.py": "",
                  "extensions/commands/cmd_foo.py": "",
                  })

    tc.run("profile detect --name=foo")
    tc.run("remote add bar http://fakeurl")

    tc.save({"conanfile.py": GenConanfile("pkg", "0.1")})
    tc.run("create .")

    assert os.path.exists(absolut_storage_path)

    tc.run("config clean")
    tc.run("profile list")
    assert "foo" not in tc.out
    tc.run("remote list")
    assert "bar" not in tc.out
    tc.run("config show core.upload:retry")
    assert "7" not in tc.out
    assert os.path.exists(os.path.join(tc.cache_folder, "extensions"))
    assert not os.path.exists(os.path.join(tc.cache_folder, "extensions", "compatibility", "mycomp.py"))
    assert os.path.exists(absolut_storage_path)
    # This will error because the call to clean will remove the profiles
    tc.run("create .", assert_error=True)
    # Works after regenerating them!
    tc.run("profile detect")
    tc.run("create .")


def test_config_reinit():
    custom_global_conf = "core.upload:retry=7"
    global_conf_folder = temp_folder()
    with open(os.path.join(global_conf_folder, "global.conf"), "w") as f:
        f.write(custom_global_conf)

    cache_folder = temp_folder()
    conan_api = ConanAPI(cache_folder=cache_folder)
    assert conan_api._api_helpers.global_conf.get("core.upload:retry", check_type=int) != 7

    conan_api.config.install(global_conf_folder, verify_ssl=False)
    # Already has an effect, the config installation reinitializes the config
    assert conan_api._api_helpers.global_conf.get("core.upload:retry", check_type=int) == 7


def test_config_reinit_core_conf():
    tc = TestClient(light=True)
    tc.save_home({"extensions/commands/cmd_foo.py": textwrap.dedent("""
        from conan.cli.command import conan_command
        from conan.api.output import ConanOutput

        @conan_command()
        def foo(conan_api, parser, *args, **kwargs):
            ''' Foo '''
            parser.parse_args(*args)
            ConanOutput().info(f"Retry: {conan_api.config.get('core.upload:retry', check_type=int)}")
    """)})
    tc.run("foo -cc core.upload:retry=7")
    assert "Retry: 7" in tc.out



import textwrap


def test_conditional_build_type(matrix_client_debug):
    # https://github.com/conan-io/conan/issues/15851
    c = matrix_client_debug
    # A header-only library can't be used for testing, it doesn't fail

    c.save({}, clean_first=True)
    c.run("new cmake_lib -d name=pkgb -d version=0.1 -d requires=matrix/1.0")
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.cmake import CMakeToolchain, CMake, cmake_layout, CMakeDeps

        class pkgbRecipe(ConanFile):
            name = "pkgb"
            version = "0.1"
            package_type = "static-library"
            settings = "os", "compiler", "build_type", "arch"
            exports_sources = "CMakeLists.txt", "src/*", "include/*"

            def layout(self):
                cmake_layout(self)

            def generate(self):
                deps = CMakeDeps(self)
                deps.generate()
                tc = CMakeToolchain(self)
                if self.settings.build_type == "Debug":
                    tc.cache_variables["USE_MATRIX"] = 1
                    tc.preprocessor_definitions["USE_MATRIX"] = 1
                tc.generate()

            def build(self):
                cmake = CMake(self)
                cmake.configure()
                cmake.build()

            def package(self):
                cmake = CMake(self)
                cmake.install()

            def package_info(self):
                self.cpp_info.libs = ["pkgb"]

            def requirements(self):
                if self.settings.build_type == "Debug":
                    self.requires("matrix/1.0")
        """)
    cmake = textwrap.dedent("""\
        set(CMAKE_CXX_COMPILER_WORKS 1)
        set(CMAKE_CXX_ABI_COMPILED 1)
        cmake_minimum_required(VERSION 3.15)
        project(pkgb CXX)

        add_library(pkgb src/pkgb.cpp)
        target_include_directories(pkgb PUBLIC include)

        if(USE_MATRIX)
            find_package(matrix CONFIG REQUIRED)
            target_link_libraries(pkgb PRIVATE matrix::matrix)
        endif()

        set_target_properties(pkgb PROPERTIES PUBLIC_HEADER "include/pkgb.h")
        install(TARGETS pkgb)
        """)
    pkgb_cpp = textwrap.dedent(r"""
        #include <iostream>
        #include "pkgb.h"
        #ifdef USE_MATRIX
        #include "matrix.h"
        #endif

        void pkgb(){
            #ifdef USE_MATRIX
            matrix();
            #endif

            #ifdef NDEBUG
            std::cout << "pkgb/0.1: Hello World Release!\n";
            #else
            std::cout << "pkgb/0.1: Hello World Debug!\n";
            #endif
        }
        """)
    c.save({"conanfile.py": conanfile,
            "CMakeLists.txt": cmake,
            "src/pkgb.cpp": pkgb_cpp})
    c.run("create . -s build_type=Debug -tf=")
    assert "matrix/1.0" in c.out
    c.run("create . -s build_type=Release -tf=")  # without dep to matrix
    assert "matrix" not in c.out

    c.save({}, clean_first=True)
    c.run("new cmake_lib -d name=pkgc -d version=0.1 -d requires=pkgb/0.1")
    c.run("build . -s build_type=Debug")
    c.run("build . -s build_type=Release")
    # This used to crash because "matrix::matrix"
    assert "conanfile.py (pkgc/0.1): Running CMake.build()" in c.out



from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient


class TestInstallMissingDependency:

    def test_missing_dep(self):
        client = TestClient()

        # Create deps packages
        dep1_conanfile = GenConanfile("dep1")
        client.save({"conanfile.py": dep1_conanfile}, clean_first=True)
        client.run("create . --name=dep1 --version=1.0 --user=lasote --channel=testing")
        client.run("create . --name=dep1 --version=2.0 --user=lasote --channel=testing")

        dep2_conanfile = GenConanfile("dep2", "1.0").with_require("dep1/1.0@lasote/testing")
        client.save({"conanfile.py": dep2_conanfile}, clean_first=True)
        client.run("create . --user=lasote --channel=testing")

        # Create final package
        # foo -------------> dep1/1.0
        #   \ -> dep2/1.0---->/
        conanfile = GenConanfile("foo", "1.0").with_require("dep1/1.0@lasote/testing")\
                                              .with_require("dep2/1.0@lasote/testing")
        client.save({"conanfile.py": conanfile}, clean_first=True)
        client.run("create . --user=lasote --channel=testing")

        # Bump version of one dependency
        # foo -------------> dep1/2.0
        #   \ -> dep2/1.0---->/
        conanfile = GenConanfile("foo", "1.0").with_requirement("dep1/2.0@lasote/testing", force=True) \
                                              .with_require("dep2/1.0@lasote/testing")
        client.save({"conanfile.py": conanfile}, clean_first=True)
        client.run("create . --user=lasote --channel=testing", assert_error=True)
        client.assert_overrides({"dep1/1.0@lasote/testing":
                                ['dep1/2.0@lasote/testing']})

        assert "Can't find a 'dep2/1.0@lasote/testing' package" in client.out
        assert "dep1/2.Y.Z" in client.out

    def test_missing_multiple_dep(self):
        client = TestClient()

        dep1_conanfile = GenConanfile()
        client.save({"conanfile.py": dep1_conanfile}, clean_first=True)
        client.run("export . --name=dep1 --version=1.0")
        client.run("export . --name=dep2 --version=1.0")

        conanfile = GenConanfile().with_require("dep1/1.0").with_require("dep2/1.0")
        client.save({"conanfile.py": conanfile}, clean_first=True)
        client.run("create . --name=pkg --version=1.0", assert_error=True)
        assert "ERROR: Missing prebuilt package for 'dep1/1.0', 'dep2/1.0'" in client.out
        assert "Try to build locally from sources using the '--build=dep1/1.0 --build=dep2/1.0'" in client.out



import textwrap

from conan.test.utils.tools import TestClient


def test_autotools_make_parameters():
    client = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.gnu import AutotoolsToolchain, Autotools

        class Conan(ConanFile):
            settings = "os"
            generators = "AutotoolsDeps", "AutotoolsToolchain", "VirtualRunEnv"

            def run(self, cmd, *args, **kwargs):
                self.output.info(f"Running {cmd}")

            def build(self):
                autotools = Autotools(self)
                autotools.make(makefile="MyMake", target="test", args=["-j4", "VERBOSE=1"])
                autotools.install(makefile="OtherMake")
        """)

    client.save({"conanfile.py": conanfile})
    client.run("build .")
    assert "make --file=MyMake test -j4 VERBOSE=1" in client.out
    assert "make --file=OtherMake install" in client.out



import re
import textwrap

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient


def test_graph_skip_build_test():
    # app -> pkg -(test)-> gtest
    #         \---(tool)-> cmake
    c = TestClient(light=True)
    pkg = textwrap.dedent("""
        from conan import ConanFile
        class Pkg(ConanFile):
            name = "pkg"
            version = "1.0"
            test_requires = "gtest/1.0"
            tool_requires = "cmake/1.0"
        """)
    c.save({"gtest/conanfile.py": GenConanfile("gtest", "1.0"),
            "cmake/conanfile.py": GenConanfile("cmake", "1.0"),
            "pkg/conanfile.py": pkg,
            "app/conanfile.py": GenConanfile("app", "1.0").with_requires("pkg/1.0")})
    c.run("create gtest")
    c.run("create cmake")
    c.run("create pkg")
    c.run("create app -c tools.graph:skip_build=True -c tools.graph:skip_test=True")
    assert "cmake" not in c.out
    assert "gtest" not in c.out
    c.run("create app -c tools.graph:skip_test=True")
    assert "WARN: experimental: Usage of 'tools.graph:skip_test'" in c.out
    assert "WARN: tools.graph:skip_test set, but tools.build:skip_test is not" in c.out
    assert "cmake" in c.out
    assert "gtest" not in c.out
    c.run("create app -c tools.graph:skip_build=True")
    assert "cmake" not in c.out
    assert "gtest" in c.out

    c.run("install app")
    assert "cmake" in c.out
    assert "gtest" in c.out

    c.run("install app -c tools.graph:skip_build=True -c tools.graph:skip_test=True")
    assert "cmake" not in c.out
    assert "gtest" not in c.out

    c.run("install app -c tools.graph:skip_build=True --build=pkg/*", assert_error=True)
    assert "ERROR: Package pkg/1.0 skipped its test/tool requires with tools.graph:skip_build, " \
           "but was marked to be built " in c.out


def test_skip():
    # https://github.com/conan-io/conan/issues/13439
    c = TestClient()
    global_conf = textwrap.dedent("""
        tools.graph:skip_test=True
        tools.build:skip_test=True
        """)
    c.save_home({"global.conf": global_conf})
    c.save({"pkga/conanfile.py": GenConanfile("pkga", "1.0.0"),
            "pkgb/conanfile.py": GenConanfile("pkgb", "1.0.0").with_test_requires("pkga/1.0.0"),
            "pkgc/conanfile.py": GenConanfile("pkgc", "1.0.0").with_test_requires("pkgb/1.0.0")})
    c.run("create pkga")
    c.run("create pkgb")

    # Always skipped
    c.run("install pkgc")
    # correct, pkga and pkgb are not in the output at all, they have been skipped
    assert "pkga" not in c.out
    assert "pkgb" not in c.out

    # not skipping test-requires
    c.run("install pkgc -c tools.graph:skip_test=False")
    # correct, pkga and pkgb are not skipped now
    assert "pkga" in c.out
    assert "pkgb" in c.out
    # but pkga binary is not really necessary
    assert re.search(r"Skipped binaries(\s*)pkga/1.0.0", c.out)

    # skipping all but the current one
    c.run("install pkgc -c &:tools.graph:skip_test=False")
    # correct, only pkga  is skipped now
    assert "pkga" not in c.out
    assert "pkgb" in c.out



import textwrap

from conan.test.utils.tools import TestClient


class TestRunner:

    def test_ignore_error(self):
        conanfile = """from conan import ConanFile
class Pkg(ConanFile):
    def source(self):
        ret = self.run("not_a_command", ignore_errors=True)
        self.output.info("RETCODE %s" % (ret!=0))
"""
        client = TestClient()
        client.save({"conanfile.py": conanfile})
        client.run("source .")
        assert "RETCODE True" in client.out

    def test_runner_capture_output(self):
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            class Pkg(ConanFile):
                def source(self):
                    self.run("echo 'hello Conan!'")
        """)
        client = TestClient()
        client.save({"conanfile.py": conanfile})
        client.run("source .")
        assert "hello Conan!" in client.out

    def test_custom_stream_error(self):
        # https://github.com/conan-io/conan/issues/7888
        conanfile = textwrap.dedent("""
            from io import StringIO
            from conan import ConanFile
            class Pkg(ConanFile):
                def source(self):
                    my_buf = StringIO()
                    self.run('echo Hello', stdout=my_buf)
                    self.output.info("Buffer got msgs {}".format(my_buf.getvalue()))
            """)
        client = TestClient()
        client.save({"conanfile.py": conanfile})
        client.run("source .")
        assert 'conanfile.py: Buffer got msgs Hello' in client.out

    def test_custom_stream_stderr(self):
        conanfile = textwrap.dedent("""
            from io import StringIO
            from conan import ConanFile
            class Pkg(ConanFile):
                def source(self):
                    my_buf = StringIO()
                    self.run('echo Hello 1>&2', stderr=my_buf)
                    self.output.info("Buffer got stderr msgs {}".format(my_buf.getvalue()))
            """)
        client = TestClient()
        client.save({"conanfile.py": conanfile})
        client.run("source .")
        assert 'conanfile.py: Buffer got stderr msgs Hello' in client.out



import os
import re
import textwrap

from jinja2 import Template, StrictUndefined

from conan.internal import check_duplicated_generator
from conan.internal.model.dependencies import get_transitive_requires
from conan.internal.model.pkg_type import PackageType
from conan.internal.util.files import save


def _relativize_path(path, start_folder):
    """
    Returns a relative path with regard to the given folder.

    :param path: absolute or relative path.
    :param start_folder: folder to start relative to.
    :return: Unix-like path relative if matches to the given pattern.
             Otherwise, it returns the original path.
    """
    if not path or not start_folder:
        return path
    path_ = path.replace("\\", "/").replace("/./", "/")
    pattern_ = start_folder.replace("\\", "/").replace("/./", "/")
    match = re.match(pattern_, path_)
    if match:
        matching = match[0]
        if path_.startswith(matching):
            path_ = path_.replace(matching, "").strip("/")
            return path_.strip("./") or "./"
    return path


class _BazelDepBuildGenerator:
    """
    This class creates the BUILD.bazel for each dependency where it's declared all the
    necessary information to load the libraries
    """
    # If both files exist, BUILD.bazel takes precedence over BUILD
    # https://bazel.build/concepts/build-files
    dep_build_filename = "BUILD.bazel"
    dep_build_template = textwrap.dedent("""\
    {% macro cc_import_macro(libs) %}
    {% for lib_info in libs %}
    cc_import(
        name = "{{ lib_info['name'] }}_precompiled",
        {% if lib_info['is_shared'] %}
        shared_library = "{{ lib_info['lib_path'] }}",
        {% else %}
        static_library = "{{ lib_info['lib_path'] }}",
        {% endif %}
        {% if lib_info['import_lib_path'] %}
        interface_library = "{{ lib_info['import_lib_path'] }}",
        {% endif %}
        {% if lib_info["linkopts"] %}
        linkopts = [
            {% for linkopt in lib_info["linkopts"] %}
            {{ linkopt }},
            {% endfor %}
        ],
        {% endif %}
    )
    {% endfor %}
    {% endmacro %}
    {% macro cc_library_macro(obj) %}
    cc_library(
        name = "{{ obj["name"] }}",
        {% if obj["headers"] %}
        hdrs = glob([
            {% for header in obj["headers"] %}
            {{ header }},
            {% endfor %}
        ],
        allow_empty = True
        ),
        {% endif %}
        {% if obj["includes"] %}
        includes = [
            {% for include in obj["includes"] %}
            {{ include }},
            {% endfor %}
        ],
        {% endif %}
        {% if obj["defines"] %}
        defines = [
            {% for define in obj["defines"] %}
            {{ define }},
            {% endfor %}
        ],
        {% endif %}
        {% if obj["linkopts"] %}
        linkopts = [
            {% for linkopt in obj["linkopts"] %}
            {{ linkopt }},
            {% endfor %}
        ],
        {% endif %}
        {% if obj["copts"] %}
        copts = [
            {% for copt in obj["copts"] %}
            {{ copt }},
            {% endfor %}
        ],
        {% endif %}
        visibility = ["//visibility:public"],
        {% if obj["libs"] or obj["dependencies"] or obj.get("component_names", []) %}
        deps = [
            # do not sort
            {% for lib in obj["libs"] %}
            ":{{ lib.name }}_precompiled",
            {% endfor %}
            {% for name in obj.get("component_names", []) %}
            ":{{ name }}",
            {% endfor %}
            {% for dep in obj["dependencies"] %}
            "{{ dep }}",
            {% endfor %}
        ],
        {% endif %}
    )
    {% endmacro %}
    {% macro filegroup_bindirs_macro(obj) %}
    {% if obj["bindirs"] %}
    filegroup(
        name = "{{ obj["name"] }}_binaries",
        srcs = glob([
            {% for bindir in obj["bindirs"] %}
            "{{ bindir }}/**",
            {% endfor %}
        ],
        allow_empty = True
        ),
        visibility = ["//visibility:public"],
    )
    {% endif %}
    {% endmacro %}
    # Components precompiled libs
    {% for component in components %}
    {{ cc_import_macro(component["libs"]) }}
    {% endfor %}
    # Root package precompiled libs
    {{ cc_import_macro(root["libs"]) }}
    # Components libraries declaration
    {% for component in components %}
    {{ cc_library_macro(component) }}
    {% endfor %}
    # Package library declaration
    {{ cc_library_macro(root) }}
    # Filegroup library declaration
    {{ filegroup_bindirs_macro(root) }}
    """)

    def __init__(self, conanfile, dep, require):
        self._conanfile = conanfile
        self._dep = dep
        self._is_build_require = require.build
        self._transitive_reqs = get_transitive_requires(self._conanfile, dep)

    @property
    def _build_file_path(self):
        """
        Returns the absolute path to the BUILD file created by Conan
        """
        folder = os.path.join(self._get_repository_name(self._dep), self.dep_build_filename)
        return folder.replace("\\", "/")

    @property
    def _absolute_build_file_path(self):
        """
        Returns the absolute path to the BUILD file created by Conan
        """
        folder = os.path.join(self._conanfile.generators_folder, self._build_file_path)
        return folder.replace("\\", "/")

    @property
    def _package_folder(self):
        """
        Returns the package folder path
        """
        # If editable, package_folder can be None
        root_folder = self._dep.recipe_folder if self._dep.package_folder is None \
            else self._dep.package_folder
        return root_folder.replace("\\", "/")

    def _get_repository_name(self, dep):
        pkg_name = dep.cpp_info.get_property("bazel_repository_name") or dep.ref.name
        return f"build-{pkg_name}" if self._is_build_require else pkg_name

    def _get_target_name(self, dep):
        pkg_name = dep.cpp_info.get_property("bazel_target_name") or dep.ref.name
        return pkg_name

    def _get_component_name(self, dep, comp_ref_name):
        pkg_name = self._get_target_name(dep)
        if comp_ref_name not in dep.cpp_info.components:
            # foo::foo might be referencing the root cppinfo
            if dep.ref.name == comp_ref_name:
                return pkg_name
            return f"{pkg_name}-{comp_ref_name}"
        comp_name = dep.cpp_info.components[comp_ref_name].get_property("bazel_target_name")
        # If user did not set bazel_target_name, let's create a component name
        # with a namespace, e.g., dep-comp1
        return comp_name or f"{pkg_name}-{comp_ref_name}"

    def _get_headers(self, cpp_info):
        return ['"{}/**"'.format(_relativize_path(path, self._package_folder))
                for path in cpp_info.includedirs]

    def _get_bindirs(self, cpp_info):
        return [_relativize_path(bindir, self._package_folder) for bindir in cpp_info.bindirs]

    def _get_includes(self, cpp_info):
        return ['"{}"'.format(_relativize_path(path, self._package_folder))
                for path in cpp_info.includedirs]

    def _get_defines(self, cpp_info):
        return ['"{}"'.format(define.replace('"', '\\' * 3 + '"'))
                for define in cpp_info.defines]

    def _get_linkopts(self, cpp_info):
        os_build = self._dep.settings_build.get_safe("os")
        link_opt = '/DEFAULTLIB:{}' if os_build == "Windows" else '-l{}'
        system_libs = [link_opt.format(lib) for lib in cpp_info.system_libs]
        shared_flags = cpp_info.sharedlinkflags + cpp_info.exelinkflags
        frameworkdirs_flags = []
        framework_flags = []
        for frw in cpp_info.frameworks:
            framework_flags.extend(["-framework", frw])
        for frw_dir in cpp_info.frameworkdirs:
            frameworkdirs_flags.extend(["-F", frw_dir.replace("\\", "/")])
        return [f'"{flag}"' for flag in (system_libs + shared_flags + framework_flags + frameworkdirs_flags)]

    def _get_copts(self, cpp_info):
        # FIXME: long discussions between copts (-Iflag) vs includes in Bazel. Not sure yet
        # includedirsflags = ['"-I{}"'.format(_relativize_path(d, package_folder_path))
        #                     for d in cpp_info.includedirs]
        cxxflags = [var.replace('"', '\\"') for var in cpp_info.cxxflags]
        cflags = [var.replace('"', '\\"') for var in cpp_info.cflags]
        return [f'"{flag}"' for flag in (cxxflags + cflags)]

    def _get_component_requirement_names(self, cpp_info):
        """
        Get all the valid names from the requirements ones given a CppInfo object.

        For instance, those requirements could be coming from:

        ```python
        from conan import ConanFile
        class PkgConan(ConanFile):
            requires = "other/1.0"

            def package_info(self):
                self.cpp_info.requires = ["other::cmp1"]

            # Or:

            def package_info(self):
                self.cpp_info.components["cmp"].requires = ["other::cmp1"]
        ```
        """
        dep_ref_name = self._dep.ref.name
        ret = []
        for req in cpp_info.requires:
            pkg_ref_name, comp_ref_name = req.split("::") if "::" in req else (dep_ref_name, req)
            prefix = ":"  # Requirements declared in the same BUILD file
            # For instance, dep == "hello/1.0" and req == "other::cmp1" -> hello != other
            if dep_ref_name != pkg_ref_name:
                try:
                    req_conanfile = self._transitive_reqs[pkg_ref_name]
                    # Requirements declared in another dependency BUILD file
                    prefix = f"@{self._get_repository_name(req_conanfile)}//:"
                except KeyError:
                    continue  # If the dependency is not in the transitive, might be skipped
            else:  # For instance, dep == "hello/1.0" and req == "hello::cmp1" -> hello == hello
                req_conanfile = self._dep
            comp_name = self._get_component_name(req_conanfile, comp_ref_name)
            dep_name = f"{prefix}{comp_name}"
            if dep_name not in ret:
                ret.append(dep_name)
        return ret

    def _get_lib_info(self, cpp_info, deduced_cpp_info, component_name=None):

        def _lib_info(lib_name, virtual_cpp_info):
            info = {
                "name": lib_name,
                "is_shared": virtual_cpp_info.type == PackageType.SHARED,
                "lib_path": _relativize_path(virtual_cpp_info.location, self._package_folder),
                "import_lib_path": _relativize_path(virtual_cpp_info.link_location, self._package_folder),
                "linkopts": []
            }
            if info['is_shared'] and info["lib_path"] and not info["lib_path"].endswith(".dll"):
                # Issue: https://github.com/conan-io/conan/issues/19190
                # Issue: https://github.com/conan-io/conan/issues/19135
                # (UNIX) Adding the rpath flag as any application could link through the library
                # which points out a symlink, but that name does not appear in the library location
                info["linkopts"] = [f'"-Wl,-rpath,{libdir}"' for libdir in cpp_info.libdirs]
            return info

        libs = cpp_info.libs
        libs_info = []
        if libs:
            if len(libs) > 1:
                for lib_name in libs:
                    name = f'_{component_name}_{lib_name}' if component_name else f'_{lib_name}'
                    virtual_cpp_info = deduced_cpp_info.components[name]
                    libs_info.append(_lib_info(lib_name, virtual_cpp_info))
            else:
                lib_name = libs[0]
                virtual_cpp_info = deduced_cpp_info.components[component_name] if component_name else deduced_cpp_info
                libs_info.append(_lib_info(lib_name, virtual_cpp_info))
        return libs_info

    def _get_build_file_context(self):
        """
        Get the whole package information

        :return: `_BazelTargetInfo` object with the package information
        """
        build_content = {"components": [], "root": {}}
        component_names = []
        deduced_cpp_info = self._dep.cpp_info.deduce_full_cpp_info(self._dep)

        if self._dep.cpp_info.has_components:
            # Loop through all the package's components
            for comp_ref_name, cmp_cpp_info in self._dep.cpp_info.get_sorted_components().items():
                # At first, let's check if we have defined some components requires, e.g., "dep::cmp1"
                comp_requires_names = self._get_component_requirement_names(cmp_cpp_info)
                comp_name = self._get_component_name(self._dep, comp_ref_name)
                component_names.append(comp_name)
                build_content["components"].append({
                    "name": comp_name,
                    "libs": self._get_lib_info(cmp_cpp_info, deduced_cpp_info, component_name=comp_ref_name),
                    "bindirs": self._get_bindirs(cmp_cpp_info),
                    "headers": self._get_headers(cmp_cpp_info),
                    "includes": self._get_includes(cmp_cpp_info),
                    "defines": self._get_defines(cmp_cpp_info),
                    "linkopts": self._get_linkopts(cmp_cpp_info),
                    "copts": self._get_copts(cmp_cpp_info),
                    "dependencies": comp_requires_names,
                })

        pkg_name = self._get_target_name(self._dep)
        # At first, let's check if we have defined some global requires, e.g., "other::cmp1"
        requires = self._get_component_requirement_names(self._dep.cpp_info)
        # If we have found some component requires it would be enough
        if not requires:
            # If no requires were found, let's try to get all the direct dependencies,
            # e.g., requires = "other_pkg/1.0"
            requires = [
                f"@{self._get_repository_name(req)}//:{self._get_target_name(req)}"
                for req in self._transitive_reqs.values()
            ]
        cpp_info = self._dep.cpp_info
        build_content["root"] = {
            "name": pkg_name,
            "libs": self._get_lib_info(cpp_info, deduced_cpp_info),
            "bindirs": self._get_bindirs(cpp_info),
            "headers": self._get_headers(cpp_info),
            "includes": self._get_includes(cpp_info),
            "defines": self._get_defines(cpp_info),
            "linkopts": self._get_linkopts(cpp_info),
            "copts": self._get_copts(cpp_info),
            "dependencies": requires,
            "component_names": component_names
        }
        return build_content

    @property
    def dep_context(self):
        """
        Return the dependency context to fill later the conan_deps_module_extension.bzl and so on.
        """
        return {
            'repository_name': self._get_repository_name(self._dep),
            'package_folder': self._package_folder,
            'package_build_file_path': self._absolute_build_file_path,
        }

    def items(self):
        template = Template(self.dep_build_template, trim_blocks=True, lstrip_blocks=True,
                            undefined=StrictUndefined)
        content = template.render(self._get_build_file_context())
        return {self._build_file_path: content}.items()


class _BazelPathsGenerator:
    """
    Bazel 6.0 needs to know all the dependencies for its current project. So, the only way
    to do that is to tell the WORKSPACE file how to load all the Conan ones. This is the goal
    of the function created by this class, the ``load_conan_dependencies`` one.

    More information:
        * https://bazel.build/reference/be/workspace#new_local_repository

    Bazel >= 7.1 needs to know all the dependencies as well, but provided via the MODULE.bazel file.
    Therefor we provide a static repository rule to load the dependencies. This rule is used by a
    module extension, passing the package path and the BUILD file path to the repository rule.
    """
    repository_filename = "dependencies.bzl"
    modules_filename = "conan_deps_module_extension.bzl"
    repository_rules_filename = "conan_deps_repo_rules.bzl"
    repository_template = textwrap.dedent("""\
        # This Bazel module should be loaded by your WORKSPACE file.
        # Add these lines to your WORKSPACE one (assuming that you're using the "bazel_layout"):
        # load("@//conan:dependencies.bzl", "load_conan_dependencies")
        # load_conan_dependencies()

        def load_conan_dependencies():
        {% for dep_info in dependencies %}
            native.new_local_repository(
                name="{{dep_info['repository_name']}}",
                path="{{dep_info['package_folder']}}",
                build_file="{{dep_info['package_build_file_path']}}",
            )
        {% endfor %}
        """)
    module_template = textwrap.dedent("""\
        # This module provides a repo for each requires-dependency in your conanfile.
        # It's generated by the BazelDeps, and should be used in your Module.bazel file.
        load(":conan_deps_repo_rules.bzl", "conan_dependency_repo")

        def _load_dependencies_impl(mctx):
        {% for dep_info in dependencies %}
            conan_dependency_repo(
                name = "{{dep_info['repository_name']}}",
                package_path = "{{dep_info['package_folder']}}",
                build_file_path = "{{dep_info['package_build_file_path']}}",
            )
        {% endfor %}

            return mctx.extension_metadata(
                # It will only warn you if any direct
                # dependency is not imported by the 'use_repo' or even it is imported
                # but not created. Notice that root_module_direct_dev_deps can not be None as we
                # are giving 'all' value to root_module_direct_deps.
                # Fix the 'use_repo' calls by running 'bazel mod tidy'
                root_module_direct_deps = 'all',
                root_module_direct_dev_deps = [],

                # Prevent writing function content to lockfiles:
                # - https://bazel.build/rules/lib/builtins/module_ctx#extension_metadata
                # Important for remote build. Actually it's not reproducible, as local paths will
                # be different on different machines. But we assume that conan works correctly here.
                # IMPORTANT: Not compatible with bazel < 7.1
                reproducible = True,
            )

        conan_extension = module_extension(
            implementation = _load_dependencies_impl,
            os_dependent = True,
            arch_dependent = True,
        )
        """)
    repository_rules_content = textwrap.dedent("""\
        # This bazel repository rule is used to load Conan dependencies into the Bazel workspace.
        # It's used by a generated module file that provides information about the conan packages.
        # Each conan package is loaded into a bazel repository rule, with having the name of the
        # package. The whole method is based on symlinks to not copy the whole package into the
        # Bazel workspace, which is expensive.
        def _conan_dependency_repo(rctx):
            package_path = rctx.workspace_root.get_child(rctx.attr.package_path)

            child_packages = package_path.readdir()
            for child in child_packages:
                rctx.symlink(child, child.basename)

            rctx.symlink(rctx.attr.build_file_path, "BUILD.bazel")

        conan_dependency_repo = repository_rule(
            implementation = _conan_dependency_repo,
            attrs = {
                "package_path": attr.string(
                    mandatory = True,
                    doc = "The path to the Conan package in conan cache.",
                ),
                "build_file_path": attr.string(
                    mandatory = True,
                    doc = "The path to the BUILD file.",
                ),
            },
        )
        """)

    @classmethod
    def items(cls, dependencies_context):
        if not dependencies_context:
            return {}
        # Bazel 6.x, but it'll likely be dropped soon
        repository_template = Template(cls.repository_template, trim_blocks=True,
                                       lstrip_blocks=True,
                                       undefined=StrictUndefined)
        content_6x = repository_template.render(dependencies=dependencies_context)
        # Bazel 7.x files
        module_template = Template(cls.module_template, trim_blocks=True, lstrip_blocks=True,
                                   undefined=StrictUndefined)
        content = module_template.render(dependencies=dependencies_context)
        return {
            cls.repository_filename: content_6x,  # bazel 6.x compatible
            cls.modules_filename: content,
            cls.repository_rules_filename: cls.repository_rules_content,
            "BUILD.bazel": "# This is an empty BUILD file."  # Bazel needs this file in each subfolder
        }.items()


class BazelDeps:

    def __init__(self, conanfile):
        """
        :param conanfile: ``< ConanFile object >`` The current recipe object. Always use ``self``.
        """
        self._conanfile = conanfile
        #: Activates the build context for the specified Conan package names.
        self.build_context_activated = []

    def _get_requirements(self, build_context_activated):
        """
        Simply save the activated requirements (host + build + test), and the deactivated ones
        """
        # All the requirements
        host_req = self._conanfile.dependencies.host
        build_req = self._conanfile.dependencies.direct_build  # tool_requires
        test_req = self._conanfile.dependencies.test

        for require, dep in list(host_req.items()) + list(build_req.items()) + list(
            test_req.items()):
            # Require is not used at the moment, but its information could be used,
            # and will be used in Conan 2.0
            # Filter the build_requires not activated with self.build_context_activated
            if require.build and dep.ref.name not in build_context_activated:
                continue
            yield require, dep

    def generate(self):
        """
        Generates all the targets <DEP>/BUILD.bazel files, a dependencies.bzl (for bazel<7), a
        conan_deps_repo_rules.bzl and a conan_deps_module_extension.bzl file (for bazel>=7.1) one in the
        build folder.

        In case of bazel < 7, it's important to highlight that the ``dependencies.bzl`` file should
        be loaded by your WORKSPACE Bazel file:

        .. code-block:: python

            load("@//[BUILD_FOLDER]:dependencies.bzl", "load_conan_dependencies")
            load_conan_dependencies()

        In case of bazel >= 7.1, the ``conan_deps_module_extension.bzl`` file should be loaded by your
        Module.bazel file, e.g. like this:

        .. code-block:: python

            load_conan_dependencies = use_extension(
                "//build:conan_deps_module_extension.bzl",
                "conan_extension"
            )
            use_repo(load_conan_dependencies, "dep-1", "dep-2", ...)
        """
        check_duplicated_generator(self, self._conanfile)
        dependencies_context = []
        for require, dep in self._get_requirements(self.build_context_activated):
            # Bazel info generator
            dep_build_generator = _BazelDepBuildGenerator(self._conanfile, dep, require)
            dependencies_context.append(dep_build_generator.dep_context)
            for name, content in dep_build_generator.items():
                save(name, content)

        for name, content in _BazelPathsGenerator.items(dependencies_context):
            save(name, content)



import copy
import time
from collections import OrderedDict

import pytest
from unittest.mock import patch

from conan.test.utils.env import environment_update
from conan.internal.errors import RecipeNotFoundException
from conan.api.model import RecipeReference
from conans.server.revision_list import RevisionList
from conan.test.utils.tools import TestServer, GenConanfile, TestClient
from conan.internal.util.files import load


def _create(c_v2, ref, conanfile=None, args=None, assert_error=False):
    conanfile = conanfile or GenConanfile()
    c_v2.save({"conanfile.py": str(conanfile)})
    r = ref
    c_v2.run(f"create . --name {r.name} --version {r.version} "
             f"--user {r.user} --channel {r.channel} {args or ''}", assert_error=assert_error)
    if not assert_error:
        pref = c_v2.created_package_reference(str(ref))
        return pref


@pytest.mark.artifactory_ready
class TestInstallingPackagesWithRevisions:

    @pytest.fixture(autouse=True)
    def setup(self):
        self.server = TestServer()
        self.server2 = TestServer()
        self.servers = OrderedDict([("default", self.server),
                                    ("remote2", self.server2)])
        self.c_v2 = TestClient(servers=self.servers, inputs=2*["admin", "password"])
        self.ref = RecipeReference.loads("lib/1.0@conan/testing")

    def recipe_revision(self, ref):
        tmp = copy.copy(ref)
        tmp.revision = None
        latest_rrev = self.c_v2.cache.get_latest_recipe_revision(tmp)
        return latest_rrev.revision

    def package_revision(self, pref):
        tmp = copy.copy(pref)
        tmp.revision = None
        latest_prev = self.c_v2.cache.get_latest_package_revision(tmp)
        return latest_prev.revision

    def test_install_binary_iterating_remotes_same_rrev(self):
        """We have two servers (remote1 and remote2), first with a recipe but the
        second one with a PREV of the binary.
        If a client installs without specifying -r remote1, it will iterate remote2 also"""
        conanfile = GenConanfile().with_package_file("file.txt", env_var="MY_VAR")
        with environment_update({"MY_VAR": "1"}):
            pref = _create(self.c_v2, self.ref, conanfile=conanfile)
        the_time = time.time()
        with patch.object(RevisionList, '_now', return_value=the_time):
            self.c_v2.run(f"upload {self.ref} -r=default -c")
        self.c_v2.run("remove {}#*:{} -c -r default".format(self.ref, pref.package_id))
        # Same RREV, different PREV
        with environment_update({"MY_VAR": "2"}):
            pref2 = _create(self.c_v2, self.ref, conanfile=conanfile)

        the_time = the_time + 10.0
        with patch.object(RevisionList, '_now', return_value=the_time):
            self.c_v2.run(f"upload {self.ref} -r=remote2 -c")
        self.c_v2.remove_all()

        assert pref.ref.revision == pref2.ref.revision

        self.c_v2.run("install --requires={}".format(self.ref))
        self.c_v2.assert_listed_require({str(self.ref): "Downloaded (default)"})
        assert "Retrieving package {} from remote 'remote2'".format(pref.package_id) in self.c_v2.out

    def test_diamond_revisions_conflict(self):
        """ If we have a diamond because of pinned conflicting revisions in the requirements,
        it gives an error"""

        # Two revisions of "lib1" to the server
        lib1 = RecipeReference.loads("lib1/1.0@conan/stable")
        lib1_pref = _create(self.c_v2, lib1)
        self.c_v2.run(f"upload {lib1} -r=default -c")
        lib1b_pref = _create(self.c_v2, lib1, conanfile=GenConanfile().with_build_msg("Rev2"))
        self.c_v2.run(f"upload {lib1} -r=default -c")

        # Lib2 depending of lib1
        self.c_v2.remove_all()
        lib2 = RecipeReference.loads("lib2/1.0@conan/stable")
        _create(self.c_v2, lib2, conanfile=GenConanfile().with_requirement(lib1_pref.ref))
        self.c_v2.run(f"upload {lib2} -r=default -c")

        # Lib3 depending of lib1b
        self.c_v2.remove_all()
        lib3 = RecipeReference.loads("lib3/1.0@conan/stable")
        _create(self.c_v2, lib3, conanfile=GenConanfile().with_requirement(lib1b_pref.ref))
        self.c_v2.run(f"upload {lib3} -r=default -c")

        # Project depending on both lib3 and lib2
        self.c_v2.remove_all()
        project = RecipeReference.loads("project/1.0@conan/stable")
        _create(self.c_v2, project,
                conanfile=GenConanfile().with_requirement(lib2).with_requirement(lib3),
                assert_error=True)
        assert "ERROR: Version conflict" in self.c_v2.out
        # assert "Different revisions of {} has been requested".format(lib1), self.c_v2.out)

    def test_alias_to_a_rrev(self):
        """ If an alias points to a RREV, it resolved that RREV and no other"""

        # Upload one revision
        pref = _create(self.c_v2, self.ref)
        self.c_v2.run(f"upload {self.ref} -r=default -c")

        # Upload other revision
        _create(self.c_v2, self.ref, conanfile=GenConanfile().with_build_msg("Build Rev 2"))
        self.c_v2.run(f"upload {self.ref} -r=default -c")
        self.c_v2.remove_all()

        # Create an alias to the first revision
        self.c_v2.alias("lib/latest@conan/stable", repr(pref.ref))
        alias_ref = RecipeReference.loads("lib/latest@conan/stable")
        exported = load(self.c_v2.get_latest_ref_layout(alias_ref).conanfile())
        assert 'alias = "{}"'.format(repr(pref.ref)) in exported

        self.c_v2.run(f"upload lib/latest@conan/stable -r=default -c")
        self.c_v2.remove_all()

        self.c_v2.run("install --requires=lib/(latest)@conan/stable")
        # Shouldn't be packages in the cache
        assert "doesn't belong to the installed recipe revision" not in self.c_v2.out

        # Read current revision
        assert pref.ref.revision == self.recipe_revision(self.ref)

    def test_revision_metadata_update_on_install(self):
        """If a clean v2 client installs a RREV/PREV from a server, it get
        the revision from upstream"""
        # Upload with v2
        pref = _create(self.c_v2, self.ref)
        self.c_v2.run(f"upload {self.ref} -r=default -c")

        # Remove all from c_v2 local
        self.c_v2.remove_all()
        assert len(self.c_v2.cache.get_recipe_revisions(self.ref)) == 0

        self.c_v2.run("install --requires={}".format(self.ref))
        local_rev = self.recipe_revision(self.ref)
        local_prev = self.package_revision(pref)
        assert local_rev == pref.ref.revision
        assert local_prev == pref.revision

    def test_revision_update_on_package_update(self):
        """
        A client v2 upload RREV with PREV1
        Another client v2 upload the same RREV with PREV2
        The first client can upgrade from the remote, only
        in the package, because the recipe is the same and it is not updated"""
        client = TestClient(servers={"default": self.server}, inputs=["admin", "password"])
        client2 = TestClient(servers={"default": self.server}, inputs=["admin", "password"])

        conanfile = GenConanfile("pkg", "0.1").with_package_file("file", env_var="MY_VAR")
        client.save({"conanfile.py": conanfile})
        with environment_update({"MY_VAR": "1"}):
            client.run("create .")
            pref = client.created_package_reference("pkg/0.1")

        time.sleep(1)

        with patch.object(RevisionList, '_now', return_value=time.time()):
            client.run(f"upload * -r=default -c")

        client2.save({"conanfile.py": conanfile})
        with environment_update({"MY_VAR": "2"}):
            client2.run("create .")
            pref2 = client2.created_package_reference("pkg/0.1")

        with patch.object(RevisionList, '_now', return_value=time.time() + 20.0):
            client2.run(f"upload * -r=default -c")

        prev1_time_remote = self.server.package_revision_time(pref)
        prev2_time_remote = self.server.package_revision_time(pref2)
        assert prev1_time_remote != prev2_time_remote  # Two package revisions

        client.run("install --requires=pkg/0.1 --update")
        client.assert_listed_require({"pkg/0.1": "Cache (Updated date) (default)"})
        assert "Retrieving package {}".format(pref.package_id) in client.out

    def test_revision_mismatch_packages_in_local(self):
        """Test adapted for the new cache: we create a revision but we export again a  recipe
        to create a new revision, then we won't have a package for the latest recipe revision
        of the cache.
        TODO: cache2.0 check this case"""
        client = self.c_v2
        pref = _create(self.c_v2, self.ref)
        ref2 = client.export(self.ref, conanfile=GenConanfile().with_build_msg("REV2"))
        # Now we have two RREVs and a PREV corresponding to the first one
        sot1 = copy.copy(pref.ref)
        sot1.revision = None
        sot2 = copy.copy(ref2)
        sot2.revision = None
        assert sot1 == sot2
        assert pref.ref.revision != ref2.revision

        # Now we try to install the self.ref, the binary is missing when using revisions
        command = "install --requires={}".format(self.ref)
        client.run(command, assert_error=True)
        assert "ERROR: Missing prebuilt package for '{}'".format(self.ref) in client.out

    def test_revision_install_explicit_mismatch_rrev(self):
        # If we have a recipe in local, but we request to install a different one with RREV
        # It fail and won't look the remotes unless --update
        client = self.c_v2
        ref = client.export(self.ref)
        command = "install --requires={}#fakerevision".format(ref)
        client.run(command, assert_error=True)
        assert "Unable to find '{}#fakerevision' in remotes".format(ref) in client.out
        command = "install --requires={}#fakerevision --update".format(ref)
        client.run(command, assert_error=True)
        assert "Unable to find '{}#fakerevision' in remotes".format(ref) in client.out

        # Now create a new revision with other client and upload it, we will request it
        new_client = TestClient(servers=self.servers, inputs=["admin", "password"])
        new_client.save({"conanfile.py": GenConanfile("lib", "1.0").with_build_msg("Rev2")})
        new_client.run("create . --user=conan --channel=testing")
        pref = new_client.created_package_reference(self.ref)
        new_client.run(f"upload * -r=default -c")

        # Repeat the install --update pointing to the new reference
        client.run("install --requires={} --update".format(repr(pref.ref)))
        client.assert_listed_require({str(self.ref): "Downloaded (default)"})

    def test_revision_mismatch_packages_remote(self):
        """If we have a recipe that doesn't match a remote recipe:
         It is not resolved in the remote."""
        _create(self.c_v2, self.ref)
        self.c_v2.run(f"upload {self.ref} -r=default -c")

        client = self.c_v2
        client.remove_all()
        client.export(self.ref, conanfile=GenConanfile().with_build_msg("REV2"))
        command = "install --requires={}".format(self.ref)

        client.run(command, assert_error=True)
        assert "Can't find a '{}' package".format(self.ref) in client.out

    def test_revision_build_requires(self):
        conanfile = GenConanfile()

        refs = []
        for _ in range(1, 4):  # create different revisions
            conanfile.with_build_msg("any change to get another rrev")
            pref = _create(self.c_v2, self.ref, conanfile=conanfile)
            self.c_v2.run(f"upload {pref.ref} -r=default -c")
            refs.append(pref.ref)
            assert refs.count(pref.ref) == 1  # make sure that all revisions are different

        client = self.c_v2  # revisions enabled
        client.remove_all()

        for ref in refs:
            command = "install --update --tool-require={}".format(repr(ref))
            client.run(command)
            assert "Downloaded recipe revision {}".format(ref.revision) in client.out


class TestRemoveWithRevisions:

    @pytest.fixture(autouse=True)
    def setup(self):
        self.server = TestServer()
        self.c_v2 = TestClient(servers={"default": self.server}, inputs=["admin", "password"])
        self.ref = RecipeReference.loads("lib/1.0@conan/testing")

    def test_remove_local_recipe(self):
        """Locally: When I remove a recipe with RREV only if the local revision matches is removed"""
        client = self.c_v2

        # If I remove the ref, the revision is gone, of course
        ref1 = client.export(self.ref)
        ref1.revision = None
        client.run("remove {} -c".format(repr(ref1)))
        assert not client.recipe_exists(self.ref)

        # If I remove a ref with a wrong revision, the revision is not removed
        ref1 = client.export(self.ref)
        fakeref = copy.copy(ref1)
        fakeref.revision = "fakerev"
        full_ref = repr(fakeref)
        client.run("remove {} -c".format(repr(fakeref)), assert_error=True)
        assert f"ERROR: Recipe revision '{full_ref}' not found" in client.out
        assert client.recipe_exists(self.ref)

    def test_remove_local_package(self):
        """Locally:
            When I remove a recipe without RREV, the package is removed.
            When I remove a recipe with RREV only if the local revision matches is removed
            When I remove a package with PREV and not RREV it raises an error
            When I remove a package with RREV and PREV only when both matches is removed"""
        client = self.c_v2

        # If I remove the ref without RREV, the packages are also removed
        pref1 = _create(self.c_v2, self.ref)
        tmp = copy.copy(pref1.ref)
        tmp.revision = None
        client.run("remove {} -c".format(repr(tmp)))
        assert not client.package_exists(pref1)

        # If I remove the ref with fake RREV, the packages are not removed
        pref1 = _create(self.c_v2, self.ref)
        fakeref = copy.copy(pref1.ref)
        fakeref.revision = "fakerev"
        str_ref = repr(fakeref)
        client.run("remove {} -c".format(repr(fakeref)), assert_error=True)
        assert client.package_exists(pref1)
        assert "Recipe revision '{}' not found".format(str_ref) in client.out

        # If I remove the ref with valid RREV, the packages are removed
        pref1 = _create(self.c_v2, self.ref)
        client.run("remove {} -c".format(repr(pref1.ref)))
        assert not client.package_exists(pref1)

        # If I remove the ref without RREV but specifying PREV it raises
        pref1 = _create(self.c_v2, self.ref)
        tmp = copy.copy(pref1.ref)
        tmp.revision = None
        command = "remove {}:{}#{} -c".format(repr(tmp), pref1.package_id, pref1.revision)
        client.run(command)
        assert not client.package_exists(pref1)

        # A wrong PREV doesn't remove the PREV
        pref1 = _create(self.c_v2, self.ref)
        command = "remove {}:{}#fakeprev -c".format(repr(pref1.ref), pref1.package_id)
        client.run(command, assert_error=True)
        assert client.package_exists(pref1)
        assert "ERROR: Package revision" in client.out

        # Everything correct, removes the unique local package revision
        pref1 = _create(self.c_v2, self.ref)
        command = "remove {}:{}#{} -c".format(repr(pref1.ref), pref1.package_id, pref1.revision)
        client.run(command)
        assert not client.package_exists(pref1)

    def test_remove_remote_recipe(self):
        """When a client removes a reference, it removes ALL revisions, no matter
        if the client is v1 or v2"""
        pref1 = _create(self.c_v2, self.ref)
        self.c_v2.run(f"upload {pref1.ref} -r=default -c")

        pref2 = _create(self.c_v2, self.ref, conanfile=GenConanfile().with_build_msg("RREV 2!"))
        self.c_v2.run(f"upload {pref2.ref} -r=default -c")

        assert pref1 != pref2

        remover_client = self.c_v2

        # Remove ref without revision in a remote
        remover_client.run("remove {} -c -r default".format(self.ref))
        assert not self.server.recipe_exists(self.ref)
        assert not self.server.recipe_exists(pref1.ref)
        assert not self.server.recipe_exists(pref2.ref)
        assert not self.server.package_exists(pref1)
        assert not self.server.package_exists(pref2)

    def test_remove_remote_recipe_revision(self):
        """If a client removes a recipe with revision:
             - If the client is v2 will remove only that revision"""
        pref1 = _create(self.c_v2, self.ref)
        self.c_v2.run(f"upload {pref1.ref} -r=default -c")

        pref2 = _create(self.c_v2, self.ref, conanfile=GenConanfile().with_build_msg("RREV 2!"))
        self.c_v2.run(f"upload {pref2.ref} -r=default -c")

        assert pref1 != pref2

        remover_client = self.c_v2

        # Remove ref without revision in a remote
        command = "remove {} -c -r default".format(repr(pref1.ref))
        remover_client.run(command)
        assert not self.server.recipe_exists(pref1.ref)
        assert self.server.recipe_exists(pref2.ref)

    def test_remove_remote_package(self):
        """When a client removes a package, without RREV, it removes the package from ALL
        RREVs"""
        pref1 = _create(self.c_v2, self.ref)
        self.c_v2.run(f"upload {pref1.ref} -r=default -c")

        pref2 = _create(self.c_v2, self.ref, conanfile=GenConanfile().with_build_msg("RREV 2!"))
        self.c_v2.run(f"upload {pref2.ref} -r=default -c")

        assert pref1.package_id == pref2.package_id
        # Both revisions exist in 2.0 cache
        assert self.c_v2.package_exists(pref1)
        assert self.c_v2.package_exists(pref2)

        remover_client = self.c_v2

        # Remove pref without RREV in a remote
        remover_client.run("remove {}#*:{} -c -r default".format(self.ref, pref2.package_id))
        assert self.server.recipe_exists(pref1.ref)
        assert self.server.recipe_exists(pref2.ref)
        assert not self.server.package_exists(pref1)
        assert not self.server.package_exists(pref2)

    def test_remove_remote_package_revision(self):
        """When a client removes a package with PREV
          (conan remove zlib/1.0@conan/stable:12312#PREV)
            - If not RREV, the client fails
            - If RREV and PREV:
                - If v1 it fails in the client (cannot transmit revisions with v1)
                - If v2 it removes only that PREV
        """
        # First RREV
        pref1 = _create(self.c_v2, self.ref)
        self.c_v2.run(f"upload {pref1.ref} -r=default -c")

        # Second RREV with two PREVS (exactly same conanfile, different package files)
        rev2_conanfile = GenConanfile().with_build_msg("RREV 2!")\
                                       .with_package_file("file", env_var="MY_VAR")
        with environment_update({"MY_VAR": "1"}):
            pref2 = _create(self.c_v2, self.ref, conanfile=rev2_conanfile)
            self.c_v2.run(f"upload {pref2.ref} -r=default -c")

        with environment_update({"MY_VAR": "2"}):
            pref2b = _create(self.c_v2, self.ref, conanfile=rev2_conanfile)
            self.c_v2.run(f"upload {pref2b.ref} -r=default -c")

        # Check created revisions
        assert pref1.package_id == pref2.package_id
        assert pref2.package_id == pref2b.package_id
        assert pref2.ref.revision == pref2b.ref.revision
        assert pref2.revision != pref2b.revision

        remover_client = self.c_v2

        # Remove PREV without RREV in a remote, the client has to fail
        command = "remove {}:{}#{} -c -r default".format(self.ref, pref2.package_id, pref2.revision)
        remover_client.run(command)

        assert self.server.recipe_exists(pref1.ref)
        assert self.server.recipe_exists(pref2.ref)
        assert self.server.recipe_exists(pref2b.ref)
        assert self.server.package_exists(pref1)
        assert self.server.package_exists(pref2b)
        assert not self.server.package_exists(pref2)

        # Try to remove a missing revision
        command = "remove {}:{}#fakerev -c -r default".format(repr(pref2.ref), pref2.package_id)
        remover_client.run(command, assert_error=True)
        fakeref = copy.copy(pref2)
        fakeref.revision = "fakerev"
        assert f"ERROR: Package revision '{fakeref.repr_notime()}' not found" in remover_client.out


class TestUploadPackagesWithRevisions:

    @pytest.fixture(autouse=True)
    def setup(self):
        self.server = TestServer()
        self.c_v2 = TestClient(servers={"default": self.server}, inputs=["admin", "password"])
        self.ref = RecipeReference.loads("lib/1.0@conan/testing")

    def test_upload_a_recipe(self):
        """If we upload a package to a server:
        Using v2 client it will upload RREV revision to the server. The rev time is NOT updated.
        """
        client = self.c_v2
        pref = _create(self.c_v2, self.ref)
        client.run(f"upload {self.ref} -r=default -c")
        revs = [r.revision for r in self.server.server_store.get_recipe_revisions_references(self.ref)]

        assert revs == [pref.ref.revision]

    def test_upload_no_overwrite_recipes(self):
        """If we upload a RREV to the server and create a new RREV in the client,
        when we upload with --no-overwrite
        Using v2 client it will warn an upload a new revision.
        """
        client = self.c_v2
        pref = _create(self.c_v2, self.ref, conanfile=GenConanfile().with_setting("os"),
                       args=" -s os=Windows")
        client.run(f"upload {self.ref} -r=default -c")

        pref2 = _create(self.c_v2, self.ref,
                        conanfile=GenConanfile().with_setting("os").with_build_msg("rev2"),
                        args=" -s os=Linux")

        assert self.server.server_store.get_last_revision(self.ref)[0] == pref.ref.revision
        client.run(f"upload {self.ref} -r=default -c")
        assert self.server.server_store.get_last_revision(self.ref)[0] == pref2.ref.revision

    def test_upload_no_overwrite_packages(self):
        """If we upload a PREV to the server and create a new PREV in the client,
        when we upload with --no-overwrite
        Using v2 client it will warn and upload a new revision.
        """
        client = self.c_v2
        conanfile = GenConanfile().with_package_file("file", env_var="MY_VAR")
        with environment_update({"MY_VAR": "1"}):
            pref = _create(self.c_v2, self.ref, conanfile=conanfile)
        client.run(f"upload {self.ref} -r=default -c")

        with environment_update({"MY_VAR": "2"}):
            pref2 = _create(self.c_v2, self.ref, conanfile=conanfile)

        assert pref.revision != pref2.revision

        assert self.server.server_store.get_last_package_revision(pref2).revision == pref.revision
        client.run(f"upload {self.ref} -r=default -c")
        assert self.server.server_store.get_last_package_revision(pref2).revision == pref2.revision


def test_server_with_only_v2_capability():
    server = TestServer(server_capabilities=[])
    c_v2 = TestClient(servers={"default": server}, inputs=["admin", "password"])
    c_v2.save({"conanfile.py": GenConanfile("pkg", "0.1")})
    c_v2.run("create .")
    c_v2.run(f"upload * -r=default -c")


class TestServerRevisionsIndexes:

    @pytest.fixture(autouse=True)
    def setup(self):
        self.server = TestServer()
        self.c_v2 = TestClient(servers={"default": self.server}, inputs=["admin", "password"])
        self.ref = RecipeReference.loads("lib/1.0@conan/testing")

    def test_rotation_deleting_recipe_revisions(self):
        """
        - If we have two RREVs in the server and we remove the first one,
        the last one is the latest
        - If we have two RREvs in the server and we remove the second one,
        the first is now the latest
        """
        ref1 = self.c_v2.export(self.ref, conanfile=GenConanfile())
        self.c_v2.run(f"upload {ref1} -r=default -c")
        assert self.server.server_store.get_last_revision(self.ref).revision == ref1.revision
        ref2 = self.c_v2.export(self.ref, conanfile=GenConanfile().with_build_msg("I'm rev2"))
        self.c_v2.run(f"upload {ref2} -r=default -c")
        assert self.server.server_store.get_last_revision(self.ref).revision == ref2.revision
        ref3 = self.c_v2.export(self.ref, conanfile=GenConanfile().with_build_msg("I'm rev3"))
        self.c_v2.run(f"upload {ref2} -r=default -c")
        assert self.server.server_store.get_last_revision(self.ref).revision == ref3.revision

        revs = [r.revision for r in self.server.server_store.get_recipe_revisions_references(self.ref)]
        assert revs == [ref3.revision, ref2.revision, ref1.revision]
        assert self.server.server_store.get_last_revision(self.ref).revision == ref3.revision

        # Delete the latest from the server
        self.c_v2.run("remove {} -r default -c".format(repr(ref3)))
        revs = [r.revision for r in self.server.server_store.get_recipe_revisions_references(self.ref)]
        assert revs == [ref2.revision, ref1.revision]
        assert self.server.server_store.get_last_revision(self.ref).revision == ref2.revision

    def test_rotation_deleting_package_revisions(self):
        """
        - If we have two PREVs in the server and we remove the first one,
        the last one is the latest
        - If we have two PREVs in the server and we remove the second one,
        the first is now the latest
        """
        conanfile = GenConanfile().with_package_file("file", env_var="MY_VAR")
        with environment_update({"MY_VAR": "1"}):
            pref1 = _create(self.c_v2, self.ref, conanfile=conanfile)
        self.c_v2.run("upload * -r=default -c")
        assert self.server.server_store.get_last_package_revision(pref1).revision == pref1.revision
        with environment_update({"MY_VAR": "2"}):
            pref2 = _create(self.c_v2, self.ref, conanfile=conanfile)
        self.c_v2.run("upload * -r=default -c")
        assert self.server.server_store.get_last_package_revision(pref1).revision == pref2.revision
        with environment_update({"MY_VAR": "3"}):
            pref3 = _create(self.c_v2, self.ref, conanfile=conanfile)
        self.c_v2.run("upload * -r=default -c")
        assert self.server.server_store.get_last_package_revision(pref1).revision == pref3.revision

        assert pref1.ref.revision == pref2.ref.revision
        assert pref2.ref.revision == pref3.ref.revision
        assert pref3.ref.revision == pref3.ref.revision

        pref = copy.copy(pref1)
        pref.revision = None
        revs = [r.revision
                for r in self.server.server_store.get_package_revisions_references(pref)]
        assert revs == [pref3.revision, pref2.revision, pref1.revision]
        assert self.server.server_store.get_last_package_revision(pref).revision == pref3.revision

        # Delete the latest from the server
        self.c_v2.run("remove {}:{}#{} -r default -c".format(repr(pref3.ref), pref3.package_id,
                                                             pref3.revision))
        revs = [r.revision
                for r in self.server.server_store.get_package_revisions_references(pref)]
        assert revs == [pref2.revision, pref1.revision]
        assert self.server.server_store.get_last_package_revision(pref).revision == pref2.revision

    def test_deleting_all_rrevs(self):
        """
        If we delete all the recipe revisions in the server. There is no latest.
        If then a client uploads a RREV it is the latest
        """
        ref1 = self.c_v2.export(self.ref, conanfile=GenConanfile())
        self.c_v2.run("upload * -r=default -c")
        ref2 = self.c_v2.export(self.ref, conanfile=GenConanfile().with_build_msg("I'm rev2"))
        self.c_v2.run("upload * -r=default -c")
        ref3 = self.c_v2.export(self.ref, conanfile=GenConanfile().with_build_msg("I'm rev3"))
        self.c_v2.run("upload * -r=default -c")

        self.c_v2.run("remove {} -r default -c".format(repr(ref1)))
        self.c_v2.run("remove {} -r default -c".format(repr(ref2)))
        self.c_v2.run("remove {} -r default -c".format(repr(ref3)))

        with pytest.raises(RecipeNotFoundException):
            self.server.server_store.get_recipe_revisions_references(self.ref)

        ref4 = self.c_v2.export(self.ref, conanfile=GenConanfile().with_build_msg("I'm rev4"))
        self.c_v2.run("upload * -r=default -c")

        revs = [r.revision for r in self.server.server_store.get_recipe_revisions_references(self.ref)]
        assert revs == [ref4.revision]

    def test_deleting_all_prevs(self):
        """
        If we delete all the package revisions in the server. There is no latest.
        If then a client uploads a RREV/PREV it is the latest
        """
        conanfile = GenConanfile().with_package_file("file", env_var="MY_VAR")
        with environment_update({"MY_VAR": "1"}):
            pref1 = _create(self.c_v2, self.ref, conanfile=conanfile)
        self.c_v2.run(f"upload {self.ref} -r=default -c")
        with environment_update({"MY_VAR": "2"}):
            pref2 = _create(self.c_v2, self.ref, conanfile=conanfile)
        self.c_v2.run(f"upload {self.ref} -r=default -c")
        with environment_update({"MY_VAR": "3"}):
            pref3 = _create(self.c_v2, self.ref, conanfile=conanfile)
        self.c_v2.run(f"upload {self.ref} -r=default -c")

        # Delete the package revisions (all of them have the same ref#rev and id)
        command = "remove {}:{}#{{}} -r default -c".format(pref3.ref.repr_notime(), pref3.package_id)
        self.c_v2.run(command.format(pref3.revision))
        self.c_v2.run(command.format(pref2.revision))
        self.c_v2.run(command.format(pref1.revision))

        with environment_update({"MY_VAR": "4"}):
            pref4 = _create(self.c_v2, self.ref, conanfile=conanfile)
        self.c_v2.run("upload {} -r default -c".format(pref4.repr_notime()))

        pref = copy.copy(pref1)
        pref.revision = None
        revs = [r.revision
                for r in self.server.server_store.get_package_revisions_references(pref)]
        assert revs == [pref4.revision]


def test_touching_other_server():
    # https://github.com/conan-io/conan/issues/9333
    servers = OrderedDict([("remote1", TestServer()),
                           ("remote2", None)])  # None server will crash if touched
    c = TestClient(servers=servers, inputs=["admin", "password"])
    c.save({"conanfile.py": GenConanfile().with_settings("os")})
    c.run("create . --name=pkg --version=0.1 --user=conan --channel=channel -s os=Windows")
    c.run("upload * -c -r=remote1")
    c.run("remove * -c")

    # This is OK, binary found
    c.run("install --requires=pkg/0.1@conan/channel -r=remote1 -s os=Windows")
    c.run("install --requires=pkg/0.1@conan/channel -r=remote1 -s os=Linux", assert_error=True)
    assert "ERROR: Missing binary: pkg/0.1@conan/channel" in c.out


@pytest.mark.artifactory_ready
def test_reupload_older_revision():
    """ upload maintains the server history
        https://github.com/conan-io/conan/issues/7331
    """
    c = TestClient(default_server_user=True)
    c.save({"conanfile.py": GenConanfile("pkg", "0.1")})
    c.run("export .")
    rrev1 = c.exported_recipe_revision()
    c.run("upload * -r=default -c")
    c.save({"conanfile.py": GenConanfile("pkg", "0.1").with_class_attribute("potato = 42")})
    c.run("export .")
    rrev2 = c.exported_recipe_revision()
    c.run("upload * -r=default -c")

    def check_order(inverse=False):
        c.run("list pkg/0.1#* -r=default")
        out = str(c.out)
        assert rrev1 in out
        assert rrev2 in out
        if inverse:
            assert out.find(rrev1) > out.find(rrev2)
        else:
            assert out.find(rrev1) < out.find(rrev2)

    check_order()

    # If we create the same older revision, and upload, still the same order
    c.save({"conanfile.py": GenConanfile("pkg", "0.1")})
    c.run("export .")
    c.run("upload * -r=default -c")
    check_order()

    # Force doesn't change it, same order
    c.run("upload * -r=default -c --force")
    check_order()

    # the only way is to remove it, then upload
    c.run(f"remove pkg/0.1#{rrev1} -r=default -c")
    c.run("upload * -r=default -c --force")
    check_order(inverse=True)


@pytest.mark.artifactory_ready
def test_reupload_older_revision_new_binaries():
    """ upload maintains the server history
        https://github.com/conan-io/conan/pull/16621
    """
    c = TestClient(default_server_user=True)
    c.save({"conanfile.py": GenConanfile("pkg", "0.1").with_settings("os")})
    c.run("create . -s os=Linux")
    rrev1 = c.exported_recipe_revision()
    c.run("upload * -r=default -c")
    c.save({"conanfile.py": GenConanfile("pkg", "0.1").with_settings("os")
                                                      .with_class_attribute("potato = 42")})
    c.run("create . -s os=Linux")
    rrev2 = c.exported_recipe_revision()
    c.run("upload * -r=default -c")

    def check_order():
        c.run("list pkg/0.1#* -r=default")
        out = str(c.out)
        assert rrev1 in out
        assert rrev2 in out
        assert out.find(rrev1) < out.find(rrev2)

    check_order()

    # If we create the same older revision, and upload, still the same order
    # c.run("remove * -c")  # Make sure no other revision
    c.save({"conanfile.py": GenConanfile("pkg", "0.1").with_settings("os")})
    c.run("create . -s os=Windows")
    rrev3 = c.exported_recipe_revision()
    assert rrev3 == rrev1
    c.run(f"upload pkg*#{rrev3} -r=default -c --force")
    check_order()



import textwrap

import pytest

from conan.test.utils.tools import TestClient


class TestBuildEnvSource:
    @pytest.fixture()
    def client(self):
        c = TestClient()
        tool = textwrap.dedent(r"""
            import os
            from conan import ConanFile
            from conan.tools.files import chdir, save

            class Tool(ConanFile):
                name = "tool"
                version = "0.1"
                def package(self):
                    with chdir(self, self.package_folder):
                        echo = f"@echo off\necho MY-TOOL! {self.name}/{self.version}!!"
                        save(self, "bin/mytool.bat", echo)
                        save(self, "bin/mytool.sh", echo)
                        os.chmod("bin/mytool.sh", 0o777)
            """)
        c.save({"conanfile.py": tool})
        c.run("create .")
        return c

    def test_source_buildenv(self, client):
        c = client
        pkg = textwrap.dedent("""
            from conan import ConanFile
            import platform

            class Pkg(ConanFile):
                name = "pkg"
                version = "0.1"
                tool_requires = "tool/0.1"

                source_buildenv = True

                def source(self):
                    cmd = "mytool.bat" if platform.system() == "Windows" else "mytool.sh"
                    self.run(cmd)
            """)
        c.save({"conanfile.py": pkg})
        c.run("create .")
        assert "MY-TOOL! tool/0.1" in c.out

        c.run("install .")  # to generate conanbuild script first, so it is available
        c.run("source .")
        assert "MY-TOOL! tool/0.1" in c.out

    def test_source_buildenv_layout(self, client):
        c = client
        pkg = textwrap.dedent("""
            from conan import ConanFile
            import platform

            class Pkg(ConanFile):
                name = "pkg"
                version = "0.1"
                tool_requires = "tool/0.1"
                settings = "build_type"

                source_buildenv = True

                def layout(self):
                    self.folders.source = "mysrc"
                    bt = self.settings.get_safe("build_type") or "Release"
                    self.folders.generators = f"mybuild{bt}"

                def source(self):
                    cmd = "mytool.bat" if platform.system() == "Windows" else "mytool.sh"
                    self.run(cmd)
            """)
        c.save({"conanfile.py": pkg})
        c.run("create .")
        assert "MY-TOOL! tool/0.1" in c.out

        c.run("install .")  # to generate conanbuild script first, so it is available
        # But they are in a different folder, user can copy them to source folder to make
        # them available to source() method. This works
        # shutil.copytree(os.path.join(c.current_folder, "mybuild"),
        #                os.path.join(c.current_folder, "mysrc"))
        # Another possibility is user directly calling the "conanbuild" script to activate
        # The current solution defines "generators" to be robust for "conan source" command
        # defaulting to "Release" config
        c.run("source .")
        assert "MY-TOOL! tool/0.1" in c.out

    def test_source_buildenv_cmake_layout(self, client):
        c = client
        pkg = textwrap.dedent("""
            import os
            from conan import ConanFile
            from conan.tools.cmake import cmake_layout
            import platform

            class Pkg(ConanFile):
                name = "pkg"
                version = "0.1"
                tool_requires = "tool/0.1"
                settings = "build_type"

                source_buildenv = True

                def layout(self):
                    cmake_layout(self)
                    bt = self.settings.get_safe("build_type") or "Release"
                    self.folders.generators = os.path.join(bt, "generators")

                def source(self):
                    cmd = "mytool.bat" if platform.system() == "Windows" else "mytool.sh"
                    self.run(cmd)
            """)
        c.save({"conanfile.py": pkg})
        c.run("create .")
        assert "MY-TOOL! tool/0.1" in c.out

        c.run("install .")
        c.run("source .")
        assert "MY-TOOL! tool/0.1" in c.out

    def test_source_buildenv_default_fail(self, client):
        c = client

        pkg = textwrap.dedent("""
            from conan import ConanFile
            import platform

            class Pkg(ConanFile):
                name = "pkg"
                version = "0.1"
                tool_requires = "tool/0.1"

                def source(self):
                    cmd = "mytool.bat" if platform.system() == "Windows" else "mytool.sh"
                    self.run(cmd)
            """)
        c.save({"conanfile.py": pkg})
        c.run("create .", assert_error=True)
        assert "ERROR: pkg/0.1: Error in source() method, line 12" in c.out

        # Local will still work, because ``install`` generates env-scripts and no layout
        c.run("install .")
        c.run("source .")
        assert "MY-TOOL! tool/0.1" in c.out



import pytest
import textwrap

from conan.test.assets.sources import gen_function_cpp
from conan.test.utils.tools import TestClient
from test.functional.toolchains.meson._base import check_binary


@pytest.mark.tool("ninja")
@pytest.mark.tool("meson")
@pytest.mark.tool("pkg_config")
class TestMeson:
    _test_package_meson_build = textwrap.dedent("""
        project('test_package', 'cpp')
        hello = dependency('hello', version : '>=0.1')
        test_package = executable('test_package', 'test_package.cpp', dependencies: hello)
        test('test package', test_package)
        """)

    _test_package_conanfile_py = textwrap.dedent("""
        import os
        from conan import ConanFile
        from conan.tools.meson import Meson, MesonToolchain


        class TestConan(ConanFile):
            settings = "os", "compiler", "build_type", "arch"
            generators = "PkgConfigDeps"

            def requirements(self):
                self.requires(self.tested_reference_str)

            def layout(self):
                self.folders.build = "build"

            def generate(self):
                tc = MesonToolchain(self)
                tc.generate()

            def build(self):
                meson = Meson(self)
                meson.configure()
                meson.build()

            def test(self):
                meson = Meson(self)
                meson.configure()
                meson.test()
        """)

    def test_reuse(self):
        t = TestClient()
        t.run("new cmake_lib -d name=hello -d version=0.1")

        test_package_cpp = gen_function_cpp(name="main", includes=["hello"], calls=["hello"])

        t.save({"test_package/conanfile.py": self._test_package_conanfile_py,
                "test_package/meson.build": self._test_package_meson_build,
                "test_package/test_package.cpp": test_package_cpp})

        t.run("create . --name=hello --version=0.1")

        check_binary(t)



import textwrap
import pytest

from conan.test.utils.tools import TestClient

new_value = "will_break_next"

consumer = textwrap.dedent("""
from conan import ConanFile
from conan.tools.cmake import CMake

class Consumer(ConanFile):
    name = "consumer"
    version = "1.0"
    settings = "os", "compiler", "build_type", "arch"
    generators = "CMakeDeps", "CMakeToolchain"
    exports_sources = ["CMakeLists.txt"]
    requires = "hello/1.0"

    def build(self):
        cmake = CMake(self)
        cmake.configure()
        cmake.build()
""")


@pytest.mark.tool("cmake", "3.27")
def test_global_alias():
    conanfile = textwrap.dedent("""
    from conan import ConanFile

    class Hello(ConanFile):
        name = "hello"
        version = "1.0"
        settings = "os", "compiler", "build_type", "arch"

        def package_info(self):
            # the default global target is "hello::hello"
            self.cpp_info.set_property("cmake_target_aliases", ["hello"])
    """)

    cmakelists = textwrap.dedent("""
    cmake_minimum_required(VERSION 3.15)
    project(test NONE)

    find_package(hello REQUIRED)
    get_target_property(_aliased_target hello ALIASED_TARGET)
    message("hello aliased target: ${_aliased_target}")
    """)

    client = TestClient()
    client.save({"conanfile.py": conanfile})
    client.run(f"create . -c tools.cmake.cmakedeps:new={new_value}")

    client.save({"conanfile.py": consumer, "CMakeLists.txt": cmakelists})
    client.run(f"create . -c tools.cmake.cmakedeps:new={new_value}")

    assert "hello aliased target: hello::hello" in client.out


@pytest.mark.tool("cmake", "3.27")
def test_component_alias():
    conanfile = textwrap.dedent("""
    from conan import ConanFile

    class Hello(ConanFile):
        name = "hello"
        version = "1.0"
        settings = "os", "compiler", "build_type", "arch"

        def package_info(self):
            self.cpp_info.components["buy"].set_property("cmake_target_aliases",
                ["hola::adios"])
    """)

    cmakelists = textwrap.dedent("""
    cmake_minimum_required(VERSION 3.15)
    project(test NONE)

    find_package(hello REQUIRED)
    get_target_property(_aliased_target hola::adios ALIASED_TARGET)
    message("hola::adios aliased target: ${_aliased_target}")
    """)

    client = TestClient()
    client.save({"conanfile.py": conanfile})
    client.run(f"create . -c tools.cmake.cmakedeps:new={new_value}")

    client.save({"conanfile.py": consumer, "CMakeLists.txt": cmakelists})
    client.run(f"create . -c tools.cmake.cmakedeps:new={new_value}")

    assert "hola::adios aliased target: hello::buy" in client.out


@pytest.mark.tool("cmake", "3.27")
def test_global_and_component_alias():
    conanfile = textwrap.dedent("""
    from conan import ConanFile

    class Hello(ConanFile):
        name = "hello"
        version = "1.0"
        settings = "os", "compiler", "build_type", "arch"

        def package_info(self):
            self.cpp_info.set_property("cmake_target_aliases", ["hola::hola"])
            self.cpp_info.components["buy"].set_property("cmake_target_aliases",
                ["hola::adios"])
    """)

    cmakelists = textwrap.dedent("""
    cmake_minimum_required(VERSION 3.15)
    project(test NONE)

    find_package(hello REQUIRED)
    get_target_property(_aliased_target hola::hola ALIASED_TARGET)
    message("hola::hola aliased target: ${_aliased_target}")
    """)

    client = TestClient()
    client.save({"conanfile.py": conanfile})
    client.run(f"create . -c tools.cmake.cmakedeps:new={new_value}")

    client.save({"conanfile.py": consumer, "CMakeLists.txt": cmakelists})
    client.run(f"create . -c tools.cmake.cmakedeps:new={new_value}")

    assert "hola::hola aliased target: hello::hello" in client.out


@pytest.mark.tool("cmake", "3.27")
def test_custom_name():
    conanfile = textwrap.dedent("""
    from conan import ConanFile

    class Hello(ConanFile):
        name = "hello"
        version = "1.0"
        settings = "os", "compiler", "build_type", "arch"

        def package_info(self):
            self.cpp_info.set_property("cmake_target_name", "ola::comprar")
            self.cpp_info.set_property("cmake_target_aliases", ["hello"])
    """)

    cmakelists = textwrap.dedent("""
    cmake_minimum_required(VERSION 3.15)
    project(test NONE)

    find_package(hello REQUIRED)
    get_target_property(_aliased_target hello ALIASED_TARGET)
    message("hello aliased target: ${_aliased_target}")
    """)

    client = TestClient()
    client.save({"conanfile.py": conanfile})
    client.run(f"create . -c tools.cmake.cmakedeps:new={new_value}")

    client.save({"conanfile.py": consumer, "CMakeLists.txt": cmakelists})
    client.run(f"create . -c tools.cmake.cmakedeps:new={new_value}")

    assert "hello aliased target: ola::comprar" in client.out


@pytest.mark.tool("cmake", "3.27")
def test_collide_component_alias():
    conanfile = textwrap.dedent("""
    from conan import ConanFile

    class Hello(ConanFile):
        name = "hello"
        version = "1.0"
        settings = "os", "compiler", "build_type", "arch"

        def package_info(self):
            self.cpp_info.components["buy"].set_property("cmake_target_aliases", ["hello::buy"])
    """)

    cmakelists = textwrap.dedent("""
    cmake_minimum_required(VERSION 3.15)
    project(test NONE)

    find_package(hello REQUIRED)
    """)

    client = TestClient()
    client.save({"conanfile.py": conanfile})
    client.run(f"create . -c tools.cmake.cmakedeps:new={new_value}")

    client.save({"conanfile.py": consumer, "CMakeLists.txt": cmakelists})
    client.run(f"create . -c tools.cmake.cmakedeps:new={new_value}", assert_error=True)

    assert "Alias 'hello::buy' already defined as a target in hello/1.0" in client.out


@pytest.mark.tool("cmake", "3.27")
def test_collide_component_alias_to_alias():
    conanfile = textwrap.dedent("""
    from conan import ConanFile

    class Hello(ConanFile):
        name = "hello"
        version = "1.0"
        settings = "os", "compiler", "build_type", "arch"

        def package_info(self):
            self.cpp_info.components["ola"].set_property("cmake_target_aliases", ["hello::foo"])
            self.cpp_info.components["buy"].set_property("cmake_target_aliases", ["hello::foo"])
    """)

    cmakelists = textwrap.dedent("""
    cmake_minimum_required(VERSION 3.15)
    project(test NONE)

    find_package(hello REQUIRED)
    """)

    client = TestClient()
    client.save({"conanfile.py": conanfile})
    client.run(f"create . -c tools.cmake.cmakedeps:new={new_value}")

    client.save({"conanfile.py": consumer, "CMakeLists.txt": cmakelists})
    client.run(f"create . -c tools.cmake.cmakedeps:new={new_value}", assert_error=True)

    assert "Alias 'hello::foo' already defined in hello/1.0" in client.out


@pytest.mark.tool("cmake", "3.27")
@pytest.mark.parametrize("root_target", ["hello::custom", None])
def test_skip_global_if_aliased(root_target):
    target_line = f'self.cpp_info.set_property("cmake_target_name", "{root_target}")' if root_target else ""
    target_name = root_target or "hello::hello"
    conanfile = textwrap.dedent(f"""
    from conan import ConanFile

    class Hello(ConanFile):
        name = "hello"
        version = "1.0"
        settings = "os", "compiler", "build_type", "arch"

        def package_info(self):
            {target_line}
            self.cpp_info.components["foo"].set_property("cmake_target_aliases", ["{target_name}"])
    """)

    cmakelists = textwrap.dedent("""
    cmake_minimum_required(VERSION 3.15)
    project(test NONE)

    find_package(hello REQUIRED)
    get_target_property(_aliased_target hello::hello ALIASED_TARGET)
    message("hello::hello aliased target: ${_aliased_target}")
    """)

    client = TestClient()
    client.save({"conanfile.py": conanfile})
    client.run(f"create . -c tools.cmake.cmakedeps:new={new_value}")

    client.save({"conanfile.py": consumer, "CMakeLists.txt": cmakelists})
    client.run(f"create . -c tools.cmake.cmakedeps:new={new_value}", assert_error=True)

    assert f"Can't define an alias '{target_name}' for the root target '{target_name}' in hello/1.0" in client.out



from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient, TestRequester


class RequesterClass(TestRequester):

    def get(self, url, headers=None, **kwargs):
        print(f"URL: {url}-HEADERS: {headers}")
        return super(RequesterClass, self).get(url, headers=headers, **kwargs)


def test_request_info_headers():
    c = TestClient(requester_class=RequesterClass, default_server_user=True)
    conanfile = GenConanfile("pkg", "0.1").with_settings('os', 'arch', 'compiler') \
                                          .with_shared_option(False)
    c.save({'conanfile.py': conanfile})
    c.run("export .")
    c.run("install --requires=pkg/0.1 -s arch=x86_64", assert_error=True)
    assert "'Conan-PkgID-Options': 'shared=False'" in c.out
    assert "'Conan-PkgID-Settings': 'arch=x86_64;" in c.out




import textwrap

import pytest

from conan.tools.meson import Meson
from conan.internal.model.conf import ConfDefinition
from conan.test.utils.mocks import ConanFileMock, MockSettings
from conan.test.utils.test_files import temp_folder
from conan.tools.meson.helpers import get_apple_subsystem, to_cstd_flag, to_cppstd_flag


def test_meson_build():
    c = ConfDefinition()
    c.loads(textwrap.dedent("""\
        tools.build:jobs=10
    """))

    settings = MockSettings({"build_type": "Release",
                             "compiler": "gcc",
                             "compiler.version": "7",
                             "os": "Linux",
                             "arch": "x86_64"})
    conanfile = ConanFileMock()
    conanfile.settings = settings
    conanfile.display_name = 'test'
    conanfile.conf = c.get_conanfile_conf(None)

    meson = Meson(conanfile)
    meson.build()

    assert '-j10' in str(conanfile.command)


@pytest.mark.parametrize("apple_sdk, subsystem", [
    ("iphoneos", "ios"),
    ("iphonesimulator", "ios-simulator"),
    ("appletvos", "tvos"),
    ("appletvsimulator", "tvos-simulator"),
    ("watchos", "watchos"),
    ("watchsimulator", "watchos-simulator"),
    (None, "macos")
])
def test_meson_subsystem_helper(apple_sdk, subsystem):
    assert get_apple_subsystem(apple_sdk) == subsystem


@pytest.mark.parametrize("cstd, expected", [
    ("gnu23", "gnu23"),
    ("23", "c23"),
    (None, None)
])
def test_meson_to_cstd_flag(cstd, expected):
    assert to_cstd_flag(ConanFileMock(), cstd) == expected


@pytest.mark.parametrize("compiler, compiler_version, cppstd, expected", [
    ("gcc", "14.0", "26", "c++26"),
    ("gcc", "14.0", "gnu26", "gnu++26"),
    ("gcc", "14.0", "gnu23", "gnu++23"),
    ("gcc", "14.0", "23", "c++23"),
    ("gcc", "15.0", "26", "c++26"),
    ("msvc", "193", "23", "vc++latest"),
    (None, None, "26", "c++26"),
    (None, None, None, None)
])
def test_meson_to_cppstd_flag(compiler, compiler_version, cppstd, expected):
    assert to_cppstd_flag(ConanFileMock(), compiler, compiler_version, cppstd) == expected


def test_meson_install_strip():
    """When the configuration `tools.build:install_strip` is set to True,
        the Meson install command should include the `--strip` option.
    """
    c = ConfDefinition()
    c.loads("tools.build:install_strip=True")

    settings = MockSettings({"build_type": "Release",
                             "compiler": "gcc",
                             "compiler.version": "7",
                             "os": "Linux",
                             "arch": "x86_64"})
    conanfile = ConanFileMock()
    conanfile.settings = settings
    conanfile.conf = c.get_conanfile_conf(None)
    conanfile.folders.generators = "."
    conanfile.folders.set_base_generators(temp_folder())
    conanfile.folders.set_base_package(temp_folder())

    meson = Meson(conanfile)
    meson.install()

    assert '--strip' in str(conanfile.command)

def test_meson_install_cli_args():
    """When the `cli_args` are provided, the Meson install command should include them.
    """
    c = ConfDefinition()

    settings = MockSettings({"build_type": "Release",
                             "compiler": "gcc",
                             "compiler.version": "7",
                             "os": "Linux",
                             "arch": "x86_64"})
    conanfile = ConanFileMock()
    conanfile.settings = settings
    conanfile.conf = c.get_conanfile_conf(None)
    conanfile.folders.generators = "."
    conanfile.folders.set_base_generators(temp_folder())
    conanfile.folders.set_base_package(temp_folder())

    meson = Meson(conanfile)
    meson.install(cli_args=["--no-rebuild"])

    assert '--no-rebuild' in str(conanfile.command)



from conan.internal.model.info import ConanInfo, RequirementsInfo, PythonRequiresInfo
from conan.internal.model.options import Options
from conan.internal.model.settings import Settings


def test_false_values_affect_none():
    """ False value do become part of conaninfo.txt and package_id
    Only "None" Python values are discarded from conaninfo.txt
    """
    reqs = RequirementsInfo({})
    build_reqs = RequirementsInfo({})
    python_reqs = PythonRequiresInfo({}, default_package_id_mode=None)
    options = Options({"shared": [True, False], "other": [None, "1"]}, {"shared": False})
    settings = Settings({"mysetting": [1, 2, 3]})
    c = ConanInfo(settings, options, reqs, build_reqs, python_reqs)
    conaninfo = c.dumps()
    assert "shared=False" in conaninfo
    assert "other" not in conaninfo
    assert "mysetting" not in conaninfo

    settings.mysetting = 1
    conaninfo = c.dumps()
    assert "mysetting=1" in conaninfo



import os
import platform
import textwrap

import pytest

from conan.test.utils.tools import TestClient


class TestMesonToolchainAndGnuFlags:

    @pytest.mark.tool("ninja")
    @pytest.mark.tool("meson")
    @pytest.mark.tool("pkg_config")
    def test_mesondeps_flags_are_being_appended_and_not_replacing_toolchain_ones(self):
        """
        Test PkgConfigDeps and MesonToolchain are keeping all the flags/definitions defined
        from both generators and nothing is being messed up.
        """
        client = TestClient(path_with_spaces=False)
        if platform.system() == "Windows":
            deps_flags = '"/GA", "/analyze:quiet"'
            flags = '"/Wall", "/W4"'
        else:
            deps_flags = '"-Wpedantic", "-Werror"'
            flags = '"-Wall", "-finline-functions"'
        # Dependency - hello/0.1
        conanfile_py = textwrap.dedent("""
        from conan import ConanFile

        class HelloConan(ConanFile):
            name = "hello"
            version = "0.1"

            def package_info(self):
                self.cpp_info.cxxflags = [{}]
                self.cpp_info.defines = ['DEF1=one_string', 'DEF2=other_string']
        """.format(deps_flags))
        client.save({"conanfile.py": conanfile_py})
        client.run("create .")
        # Dependency - other/0.1
        conanfile_py = textwrap.dedent("""
        from conan import ConanFile

        class OtherConan(ConanFile):
            name = "other"
            version = "0.1"

            def package_info(self):
                self.cpp_info.defines = ['DEF3=simple_string']
        """)
        client.save({"conanfile.py": conanfile_py}, clean_first=True)
        client.run("create .")

        # Consumer using PkgConfigDeps and MesonToolchain
        conanfile_py = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.meson import Meson, MesonToolchain
        from conan.tools.gnu import PkgConfigDeps

        class App(ConanFile):
            settings = "os", "arch", "compiler", "build_type"
            requires = "hello/0.1", "other/0.1"

            def layout(self):
                self.folders.build = "build"

            def generate(self):
                deps = PkgConfigDeps(self)
                deps.generate()
                tc = MesonToolchain(self)
                tc.preprocessor_definitions["VAR"] = "VALUE"
                tc.preprocessor_definitions["VAR2"] = "VALUE2"
                tc.generate()

            def build(self):
                meson = Meson(self)
                meson.configure()
                meson.build()
        """)

        meson_build = textwrap.dedent("""
            project('tutorial', 'cpp')
            cxx = meson.get_compiler('cpp')
            hello = dependency('hello', version : '>=0.1')
            other = dependency('other', version : '>=0.1')
            # It's not needed to declare "hello/0.1" as a dependency, only interested in flags
            executable('demo', 'main.cpp', dependencies: [hello, other])
        """)

        main = textwrap.dedent("""
            #include <stdio.h>
            #define STR(x)   #x
            #define SHOW_DEFINE(x) printf("%s=%s", #x, STR(x))
            int main(int argc, char *argv[]) {
                SHOW_DEFINE(VAR);
                SHOW_DEFINE(VAR2);
                SHOW_DEFINE(DEF1);
                SHOW_DEFINE(DEF2);
                SHOW_DEFINE(DEF3);
                return 0;
            }
        """)

        client.save({"conanfile.py": conanfile_py,
                     "meson.build": meson_build,
                     "main.cpp": main},
                    clean_first=True)

        client.run("build . -c 'tools.build:cxxflags=[%s]'" % flags)

        app_name = "demo.exe" if platform.system() == "Windows" else "demo"
        client.run_command(os.path.join("build", app_name))
        assert 'VAR="VALUE' in client.out
        assert 'VAR2="VALUE2"' in client.out
        assert 'DEF1=one_string' in client.out
        assert 'DEF2=other_string' in client.out
        assert 'DEF3=simple_string' in client.out



import textwrap

import pytest
import platform

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient


executable = "myapp.bat" if platform.system() == "Windows" else "myapp.sh"


@pytest.fixture(scope="module")
def client():
    tc = TestClient(default_server_user=True)
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.files import save
        import os

        class Pkg(ConanFile):
            name = "pkg"
            version = "0.1"
            # So that the requirement is run=True even for --requires
            package_type = "application"
            options = {"foo": [True, False, "bar"]}
            default_options = {"foo": True}
            settings = "os"

            def package(self):
                executable = os.path.join(self.package_folder, "bin", '""" + executable + """')
                save(self, executable, f"echo Hello World! foo={self.options.foo}")
                # Make it executable
                os.chmod(executable, 0o755)
        """)
    tc.save({"pkg/conanfile.py": conanfile})
    tc.run("create pkg")
    return tc


@pytest.mark.parametrize("context_flag", ["host", "build", None])
@pytest.mark.parametrize("requires_context", ["host", "build",])
@pytest.mark.parametrize("use_conanfile", [True, False])
def test_run(client, context_flag, requires_context, use_conanfile):
    context_arg = {
        "host": "--context=host",
        "build": "--context=build",
        None: "",
    }.get(context_flag)
    should_find_binary = (context_flag == requires_context) or (context_flag is None)
    if use_conanfile:
        conanfile_consumer = GenConanfile("consumer", "1.0").with_settings("os")
        if requires_context == "host":
            conanfile_consumer.with_requires("pkg/0.1")
        else:
            conanfile_consumer.with_tool_requires("pkg/0.1")

        client.save({"conanfile.py": conanfile_consumer})
        client.run(f"run {executable} {context_arg}", assert_error=not should_find_binary)
    else:
        requires = "requires" if requires_context == "host" else "tool-requires"
        client.run(f"run {executable} --{requires}=pkg/0.1 {context_arg}",
                   assert_error=not should_find_binary)
    if should_find_binary:
        assert "Hello World!" in client.out
    else:
        if platform.system() == "Windows":
            assert "not recognized as an internal or external command" in client.out
        else:
            assert "Error 127 while executing" in client.out


def test_run_context_priority(client):
    client.run("create pkg -o=pkg/*:foo=False")

    client.run(f"run {executable} --requires=pkg/0.1 --tool-requires=pkg/0.1 -o:b=pkg/*:foo=False")
    # True is host, False is build, run gives priority to host
    assert "Hello World! foo=True" in client.out


def test_run_missing_executable(client):
    client.run(f"run a-binary-name-that-does-not-exist --requires=pkg/0.1", assert_error=True)
    if platform.system() == "Windows":
        assert "not recognized as an internal or external command" in client.out
    else:
        assert "Error 127 while executing" in client.out


def test_run_missing_binary(client):
    client.run("run foo --requires=pkg/0.1 -o=pkg/*:foo=bar", assert_error=True)
    assert "Error installing the dependencies" in client.out
    assert "Missing prebuilt package for 'pkg/0.1'" in client.out


def test_run_missing_package(client):
    client.run("run foo --requires=pkg/2.1", assert_error=True)
    assert "Error installing the dependencies" in client.out
    assert "Package 'pkg/2.1' not resolved" in client.out


@pytest.mark.skipif(platform.system() == "Windows", reason="Unix only")
def test_run_status_is_propagated(client):
    client.run("run false --requires=pkg/0.1", assert_error=True)
    assert "Error installing the dependencies" not in client.out
    assert "ERROR: Error 1 while executing" in client.out
